# -*- coding: utf-8 -*-
"""Bases for MOF ML model benchmarking."""
import base64
import hashlib
import os
import time
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import numpy as np
from loguru import logger
from pydantic import BaseModel, Field

from mofdscribe.bench.watermark import get_watermark
from mofdscribe.datasets.dataset import AbstractStructureDataset
from mofdscribe.metrics.metric_collection import RegressionMetricCollection
from mofdscribe.metrics.regression import get_regression_metrics
from mofdscribe.splitters.splitters import BaseSplitter
from mofdscribe.version import get_version

__all__ = ["MOFBenchRegression", "BenchResult"]

_RST_TEMPLATE = """{modelname}
------------------------------------

Model card
..............

Feature set description
~~~~~~~~~~~~~~~~~~~~~~~~~~

What features are used?
#######################

Why are the features informative?
###################################


Why do the features not leak information?
##############################################


Will the features be accessible in real-world test applications?
###################################################################

Data split
~~~~~~~~~~

Describe preprocessing steps and how data leakage is avoided
##############################################################

Describe the feature selection steps and how data leakage is avoided
#####################################################################


Describe the model selection steps and how data leakage is avoided
#####################################################################

"""

_CITATION_TEMPLATE = """"
Reference
..............

.. code::

    {bibtex}
"""


class BenchTaskEnum(Enum):
    """Enum for benchmark tasks."""

    logKH_CO2_id = "logKH_CO2_id"  # noqa: N815
    logKH_CO2_ood = "logKH_CO2_ood"  # noqa: N815
    pbe_bandgap_id = "pbe_bandgap_id"  # noqa: N815
    ch4dc_id = "ch4dc_id"  # noqa: N815


class BenchResult(BaseModel):
    """Model for benchmark results."""

    start_time: datetime = Field(title="Start time", description="Start time of the benchmark")
    end_time: datetime = Field(title="End time", description="End time of the benchmark")
    # ToDo: Add ClassificationMetricCollection
    metrics: RegressionMetricCollection = Field(
        title="Metrics", description="Metrics of the benchmark"
    )
    version: Optional[str] = Field(title="Version", description="Version of the benchmark")
    features: Optional[str] = Field(
        title="Features",
        description="Features used in the model. If you use a graph net, report the edge and vertex features.",
    )
    name: str = Field(title="Name", description="Name of the model. This will be used as filename.")
    task: BenchTaskEnum = Field(title="Task", description="Task of the benchmark")
    model_type: Optional[str] = Field(
        title="Model type", description="Model type, e.g. 'CGCNN', 'BERT', 'XGBoost'"
    )
    reference: Optional[str] = Field(title="Reference", description="Reference with more details")
    implementation: Optional[str] = Field(
        title="Implementation", description="Link to implementation"
    )
    mofdscribe_version: str = Field(title="mofdscribe version", description="mofdscribe version")
    session_info: Dict[str, Any] = Field(
        title="Session info",
        description="Automatically captured string describing the computational environment.",
    )
    logs: Optional[List[Dict]] = Field(
        title="Logs",
        description="Additional data logged using the log method during the benchmark.",
    )

    def save_json(self, folder: Union[str, os.PathLike]) -> None:
        """Save benchmark results to json file."""
        if not os.path.exists(folder):
            os.makedirs(folder)
        with open(
            os.path.join(
                folder,
                f"{id_for_bench_result(self)}.json",
            ),
            "w",
        ) as handle:
            handle.write(self.json())

    def save_rst(self, folder: Union[str, os.PathLike]) -> None:
        """Prepare RST file for model card and free-text model description."""
        if not os.path.exists(folder):
            os.makedirs(folder)
        with open(
            os.path.join(
                folder,
                f"{id_for_bench_result(self)}.rst",
            ),
            "w",
        ) as handle:
            text = _RST_TEMPLATE.format(modelname=self.name)
            if self.reference is not None:
                text += _CITATION_TEMPLATE.format(bibtex=self.reference)
            handle.write(text)


def id_for_bench_result(bench_result: BenchResult) -> str:
    """Generate an ID for a benchmark result."""
    hasher = hashlib.sha1(str(bench_result).encode("utf8"))  # noqa: S303 ok to use unsafe hash here
    hash = base64.urlsafe_b64encode(hasher.digest()).decode("ascii")

    n = bench_result.name[:2] if bench_result.name else hash[:2]
    f = bench_result.features[:2] if bench_result.features else hash[2:4]
    m = bench_result.model_type[:2] if bench_result.model_type else hash[4:6]
    r = bench_result.reference[:2] if bench_result.reference else hash[6:8]
    datetime_part = bench_result.start_time.strftime("%Y%m%d%H%M%S")
    return f"R{n}{f}{m}{r}{datetime_part}".replace("/", "-").replace(" ", "_")


class MOFBench(ABC):
    """Base class for MOFBench.

    MOFBench is a class that encapsulates the benchmarking of a model.
    It is meant to be used as a base class for other classes that implement
    task specific benchmarking.

    Each task-specific benchmarking class will be specific to a dataset
    and a splitting technique. We will not define feature vectors in the
    benchmarking classes but only indices, structures and labels.

    The benchmarking class will then call the train and predict function of a model
    multiple times and then produce a report.
    With this report, users can make a PR to be added to the leaderboard
    (for which a template exists in which they are also asked to provide a link
    to an archived version of their model code).

    The model class can be as simple as in the following example

    .. code-block:: python

        class MyDummyModel:
            def __init__(self):
                self.model = Pipeline(
                    [("scaler", StandardScaler()), ("lr", LinearRegression(**(lr_kwargs or {})))]
                )

            def featurize(self, s: Structure):
                return s.density

            def fit(self, idx, structures, y):
                x = np.array([self.featurize(s) for s in structures]).reshape(-1, 1)
                self.model.fit(x, y)

            def predict(self, idx, structures):
                x = np.array([self.featurize(s) for s in structures]).reshape(-1, 1)
                return self.model.predict(x)

    Of course, instead of re-featurizing every structure you might also
    choose to use the indices to look up pre-computed features.

    ..warning::

        This class will monkey path a :code:`log` method into the model object.
    """

    def __init__(
        self,
        model,
        ds: AbstractStructureDataset,
        splitter: BaseSplitter,
        target: List[str],
        name: str,
        task: BenchTaskEnum,
        k: int = 5,
        model_type: Optional[str] = None,
        version: Optional[str] = None,
        features: Optional[str] = None,
        reference: Optional[str] = None,
        implementation: Optional[str] = None,
        debug: bool = False,
        patch_in_ds: bool = False,
    ):
        """Initialize the benchmarking class.

        Args:
            model: Model to be benchmarked.
            ds (AbstractStructureDataset): Dataset to be used for benchmarking.
            splitter (BaseSplitter): Splitter to be used for benchmarking.
            target (List[str]): Target labels to be used for benchmarking.
                Must be included in the dataset.
            name (str): Name of the model. This will be used as filename.
            task (BenchTaskEnum): Task of the benchmark.
            k (int): Number of folds for k-fold cross-validation.
            model_type (str, optional): Model type, e.g. 'CGCNN', 'BERT', 'XGBoost'.
                Defaults to None.
            version (str, optional): Version of the model. Defaults to None.
            features (str, optional): Features used in the model. If you use a graph net,
                report the edge and vertex features. Defaults to None.
            reference (str, optional): Reference with more details
                Please provide it in BibTeX form. Defaults to None.
            implementation (str, optional): Link to implementation. Defaults to None.
            debug (bool): If True, the benchmark will be run in debug mode
                (1% of the data).
            patch_in_ds (bool): If True, the dataset will be patched into the model class
                under the `ds` attribute.
        """
        self._model = model
        self._start_time = None
        self._end_time = None
        self._fitted = False
        self._ds = ds
        self._splitter = splitter
        self._task = task
        self._version = version
        self._features = features
        self._name = name
        self._model_type = model_type
        self._reference = reference
        self._implementation = implementation
        self._debug = debug
        self._targets = target
        self._k = k
        if patch_in_ds:
            self._model.ds = self._ds
        # mokey patch for logging method
        self._model.log = self.log
        self.logs = []

    def _train(self, idx: np.ndarray, structures: np.ndarray, y: np.ndarray):
        self._model.fit(idx, structures, y)
        self._fitted = True

    def _predict(self, idx: np.ndarray, structures: np.ndarray):
        return self._model.predict(idx, structures)

    def log(self, data: dict):
        """Log data to the logs list.

        Args:
            data (dict): Data to be logged.
        """
        self.logs.append(data)

    @abstractmethod
    def _score(self):
        raise NotImplementedError

    def bench(self) -> BenchResult:
        """Run the benchmarking."""
        start_time = time.time()
        metrics = self._score()
        end_time = time.time()
        return BenchResult(
            start_time=start_time,
            end_time=end_time,
            metrics=metrics,
            version=self._version,
            features=self._features,
            name=self._name,
            model_type=self._model_type,
            task=self._task,
            reference=self._reference,
            implementation=self._implementation,
            mofdscribe_version=get_version(),
            session_info=get_watermark(),
            logs=self.logs,
        )


class MOFBenchRegression(MOFBench):
    """Regression benchmarking class."""

    def _score(self):
        metric_collection = []
        timings = []
        inference_times = []

        for i, (train_idx, test_idx) in enumerate(
            self._splitter.k_fold(
                self._k,
            )
        ):
            logger.debug(
                f"K-fold round {i}, {len(train_idx)} train points, {len(test_idx)} test points"
            )
            start_time = time.time()
            self._train(
                train_idx,
                self._ds.get_structures(train_idx),
                self._ds._df[self._targets].iloc[train_idx].values,
            )
            end_time = time.time()
            timings.append(end_time - start_time)
            start_time = time.time()
            y_pred = self._predict(test_idx, self._ds.get_structures(test_idx))
            end_time = time.time()
            inference_times.append(end_time - start_time)
            y_true = self._ds._df[self._targets].iloc[test_idx].values
            metrics = get_regression_metrics(y_pred, y_true)
            metric_collection.append(metrics)
        return RegressionMetricCollection(
            regression_metrics=metric_collection,
            fit_timings=timings,
            inference_timings=inference_times,
        )



from mofdscribe.bench.watermark import get_watermark


def test_get_watermark():
    watermark = get_watermark()
    assert isinstance(watermark, dict)
    assert watermark["packages"] is not None
    assert watermark["system"] is not None



# -*- coding: utf-8 -*-
"""Base class for datasets."""
from typing import Collection, List, Optional, Tuple

import numpy as np
from joblib import Parallel, delayed
from loguru import logger
from pymatgen.core import IStructure, Structure

from mofdscribe.datasets.utils import (
    get_decorated_graph_hash_cached,
    get_decorated_scaffold_hash_cached,
    get_undecorated_graph_hash_cached,
)

__all__ = ["AbstractStructureDataset"]


class AbstractStructureDataset:
    """Base class for datasets."""

    def __init__(self):
        """Initialize a dataset."""
        self._structures = None
        self._target = None

        self._years = None
        self._labels = None
        self._decorated_graph_hashes = None
        self._undecorated_graph_hashes = None
        self._decorated_scaffold_hashes = None
        self._undecorated_scaffold_hashes = None
        self._densities = None

    def get_subset(self, indices: Collection[int]) -> "AbstractStructureDataset":
        raise NotImplementedError()

    @property
    def available_info(self) -> List[str]:
        raise NotImplementedError()

    @property
    def available_features(self) -> List[str]:
        raise NotImplementedError()

    @property
    def available_labels(self) -> List[str]:
        raise NotImplementedError()

    def __len__(self):
        """Return the number of structures."""
        return len(self._structures)

    def __iter__(self):
        """Iterate over the structures."""
        return self.get_structures(range(len(self)))

    def get_labels(
        self, idx: Collection[int], labelnames: Optional[Collection[str]] = None
    ) -> np.ndarray:
        raise NotImplementedError()

    def get_years(self, idx: int) -> int:
        if self._years is None:
            raise ValueError("Years are not available.")
        return self._years[idx]

    # ToDo: think about how we can cache this in memory
    def get_structures(self, idx: Collection[int]) -> Collection[Structure]:
        return Parallel(n_jobs=-1)(delayed(IStructure.from_file)(self._structures[i]) for i in idx)

    def get_filenames(self, idx: Collection[int]) -> List[Structure]:
        return [self._structures[i] for i in idx]

    # ToDo: parallelize hash computation
    def get_decorated_graph_hashes(self, idx: Collection[int]) -> str:
        if self._decorated_graph_hashes is None:
            logger.info("Computing hashes, this can take a while.")
            # fixme: this is wrong, we need to compute it for all and not only for the indexed ones
            hashes = np.array(
                Parallel(n_jobs=-1)(
                    delayed(get_decorated_graph_hash_cached)(s)
                    for s in self.get_structures(range(len(self)))
                )
            )
            self._decorated_graph_hashes = hashes
        return self._decorated_graph_hashes[idx]

    def get_undecorated_graph_hashes(self, idx: Collection[int]) -> str:
        if self._undecorated_graph_hashes is None:
            logger.info("Computing hashes, this can take a while.")
            hashes = np.array(
                Parallel(n_jobs=-1)(
                    delayed(get_undecorated_graph_hash_cached)(s)
                    for s in self.get_structures(range(len(self)))
                )
            )
            self._undecorated_graph_hashes = hashes
        return self._undecorated_graph_hashes

    def get_decorated_scaffold_hashes(self, idx: Collection[int]) -> str:
        if self._decorated_graph_hashes is None:
            logger.info("Computing hashes, this can take a while.")
            hashes = np.array(
                Parallel(n_jobs=-1)(
                    delayed(get_decorated_scaffold_hash_cached)(s)
                    for s in self.get_structures(range(len(self)))
                )
            )
            self._decorated_scaffold_hashes = hashes

        return self._decorated_scaffold_hashes[idx]

    def get_undecorated_scaffold_hashes(self, idx: Collection[int]) -> str:
        if self._undecorated_scaffold_hashes is None:
            logger.info("Computing hashes, this can take a while.")
            hashes = np.array(
                Parallel(n_jobs=-1)(
                    delayed(get_undecorated_graph_hash_cached)(s)
                    for s in self.get_structures(range(len(self)))
                )
            )
            self._undecorated_scaffold_hashes = hashes
        return self._undecorated_scaffold_hashes[idx]

    def get_densities(self, idx: Collection[int]) -> np.ndarray:
        if self._densities is None:

            def get_denstiy(s):
                return s.density

            self._densities = np.array(
                Parallel(n_jobs=-1)(
                    delayed(get_denstiy)(s) for s in self.get_structures(range(len(self)))
                )
            )
        return self._densities[idx]

    # ToDo: think how this should behave.
    def select(self, indices: Collection[int], labels: Optional[Collection[str]] = None):
        return indices

    def show_structure(self, index):
        import nglview as nv

        structure = list(self.get_structures(index))[0]
        return nv.show_pymatgen(structure)

    @property
    def citations(self) -> Tuple[str]:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""RASPA force-field builder taken from aiida-lsmo.

MIT License

Copyright 2018-2021, LABORATORY OF MOLECULAR SIMULATION (LSMO),
ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

#ToDo: move this to a standalone-package that can be imported here and in aiida-lsmo
"""
import os
from math import sqrt
from typing import List

import ruamel.yaml as yaml

THISDIR = os.path.dirname(os.path.abspath(__file__))


def check_ff_list(inp_list: List[str]):
    """Check a list of atom types.

    1) Remove duplicates, preserving the order of the elements.
    2) Warn if there are atom types with the same name but different parameters
    3) If a shorter atom type comes later, swap the order # TODO!

    Args:
        inp_list (List[str]): list of atom types

    Raises:
        ValueError: If the list contains duplicate atom type with different parameters

    Returns:
        list[str]: list of atom types
    """ """ """
    out_list = []
    for item in inp_list:
        if item.split()[0] not in [x.split()[0] for x in out_list]:  # atom type label is unique
            out_list.append(item)
        else:
            if item in out_list:  # Two atom types are exactly the same
                pass
            else:
                raise ValueError("Two atom types with same name but different parameters are used.")
    return out_list


def load_yaml():
    """Load the ff_data.yaml as a dict."""
    yamlfullpath = os.path.join(THISDIR, "ff_data.yaml")

    with open(yamlfullpath, "r") as stream:
        ff_data = yaml.safe_load(stream)

    return ff_data


def render_ff_mixing_def(ff_data, params):
    """Render the force_field_mixing_rules.def file."""
    output = [
        "# general rule for shifted vs truncated (file generated by aiida-raspa)",
        ["truncated", "shifted"][params["shifted"]],
        "# general rule tail corrections",
        ["no", "yes"][params["tail_corrections"]],
        "# number of defined interactions",
    ]

    force_field_lines = []
    ff_mix_found = False  # If becomes True, needs to handle the mixing differently
    # TODO: this needs to be sorted for python versions where dictionaries are not sorted! #pylint: disable=fixme
    # If separate_interactions==True, prints only "none" interactions for the molecules
    for atom_type, ff_pot in ff_data["framework"][params["ff_framework"]]["atom_types"].items():
        force_field_lines.append(" ".join([str(x) for x in [atom_type] + ff_pot]))
    for molecule, ff_name in params["ff_molecules"].items():
        for atom_type, val in ff_data[molecule][ff_name]["atom_types"].items():
            if "force_field_mix" in val:
                ff_mix_found = True
                ff_pot = val["force_field_mix"]
            elif val["force_field"] == "use-framework-ff":
                continue
            else:
                ff_pot = val["force_field"]
            # In case of "separate_interactions" write the ff only if none particle
            if not params["separate_interactions"] or ff_pot[0].lower() == "none":
                force_field_lines.append(" ".join([str(x) for x in [atom_type] + ff_pot]))

    force_field_lines = check_ff_list(force_field_lines)
    output.append(len(force_field_lines))
    output.append("# atom_type, interaction, parameters")
    output.extend(force_field_lines)
    output.append("# general mixing rule for Lennard-Jones")
    output.append(params["mixing_rule"])
    string = "\n".join([str(x) for x in output]) + "\n"
    return string, ff_mix_found


def mix_molecule_ff(ff_list, mixing_rule):
    """Mix molecule-molecule interactions in case of separate_interactions: return mixed ff_list."""
    ff_mix = []
    for i, ffi in enumerate(ff_list):
        for ffj in ff_list[i:]:
            if ffi[1].lower() == ffj[1].lower() == "lennard-jones":
                eps_mix = sqrt(ffi[2] * ffj[2])
                if mixing_rule == "lorentz-berthelot":
                    sig_mix = 0.5 * (ffi[3] + ffj[3])
                elif mixing_rule == "jorgensen":
                    sig_mix = sqrt(ffi[3] * ffj[3])
                ff_mix.append(
                    "{} {} lennard-jones {:.5f} {:.5f}".format(ffi[0], ffj[0], eps_mix, sig_mix)
                )
            elif "none" in [ffi[1], ffj[1]]:
                ff_mix.append("{} {} none".format(ffi[0], ffj[0]))
            elif ffi[1].lower() == ffj[1].lower() == "feynman-hibbs-lennard-jones":
                eps_mix = sqrt(ffi[2] * ffj[2])
                if mixing_rule == "lorentz-berthelot":
                    sig_mix = 0.5 * (ffi[3] + ffj[3])
                elif mixing_rule == "jorgensen":
                    sig_mix = sqrt(ffi[3] * ffj[3])
                reduced_mass = ffi[4]  # assuming that ffi==ffj, for the moment
                ff_mix.append(
                    "{} {} feynman-hibbs-lennard-jones {:.5f} {:.5f} {:.5f}".format(
                        ffi[0], ffj[0], eps_mix, sig_mix, reduced_mass
                    )
                )
            else:
                raise NotImplementedError(
                    "FFBuilder is not able to mix different/unknown potentials."
                )
    return ff_mix


def render_ff_def(ff_data, params, ff_mix_found):
    """Render the force_field.def file."""
    output = [
        "# rules to overwrite (file generated by aiida-raspa)",
        0,
        "# number of defined interactions",
        "# mixing rules to overwrite",
        0,
    ]
    if params["separate_interactions"] or ff_mix_found:
        ff_list = []
        for molecule, ff_name in params["ff_molecules"].items():
            for atom_type, val in ff_data[molecule][ff_name]["atom_types"].items():
                ff_pot = val["force_field"]
                if ff_pot == "dummy_separate":  # Exclude molatoms-moldummy interactions
                    ff_list.append([atom_type] + ["none"])
                else:
                    ff_list.append([atom_type] + ff_pot)
        mixing_rule = params["mixing_rule"].lower()
        ff_mix = mix_molecule_ff(ff_list, mixing_rule)
        output.append(len(ff_mix))
        output.append("# type1 type2 interaction")
        output.extend(ff_mix)
    else:
        output.append(0)
    string = "\n".join([str(x) for x in output]) + "\n"
    return string  # , "force_field.def"


def render_pseudo_atoms_def(ff_data, params):
    """Render the pseudo_atoms.def file."""
    output = ["# number of pseudo atoms"]

    pseudo_atoms_lines = []
    for molecule, ff_name in params["ff_molecules"].items():
        for atom_type, val in ff_data[molecule][ff_name]["atom_types"].items():
            type_settings = val["pseudo_atom"]
            pseudo_atoms_lines.append(" ".join([str(x) for x in [atom_type] + type_settings]))

    pseudo_atoms_lines = check_ff_list(pseudo_atoms_lines)
    output.append(len(pseudo_atoms_lines))
    output.append(
        "#type print as chem oxidation mass charge polarization B-factor radii connectivity "
        + "anisotropic anisotropic-type tinker-type"
    )
    output.extend(pseudo_atoms_lines)
    string = "\n".join([str(x) for x in output]) + "\n"
    return string  # , "pseudo_atoms.def"


def render_molecule_def(ff_data, params, molecule_name):
    """Render the molecule.def file containing the thermophysical data, geometry and intramolecular force field."""
    ff_name = params["ff_molecules"][molecule_name]
    ff_dict = ff_data[molecule_name][ff_name]

    output = [
        "# critical constants: Temperature [T], Pressure [Pa], and Acentric factor [-] "
        + "(file generated by aiida-raspa)",
        ff_data[molecule_name]["critical_constants"]["tc"],
        ff_data[molecule_name]["critical_constants"]["pc"],
        ff_data[molecule_name]["critical_constants"]["af"],
    ]
    if ff_dict["atomic_positions"] == "flexible":  # read intermolecular forcefield from file
        ff_intermol_path = os.path.join(
            THISDIR, "ff_flexible", "{}_{}.def".format(molecule_name, ff_name)
        )
        with open(ff_intermol_path, "r") as ff_intermol:
            output += [line.strip() for line in ff_intermol.readlines()]
    else:  # rigid molecules: atomic positions provided as list of coordinates
        natoms = len(ff_dict["atomic_positions"])
        output.append("# Number Of atoms")
        output.append(natoms)
        output.append("# Number of groups (only whole molecule)")
        output.append(1)
        output.append("# Group-1: rigid/flexible")
        output.append("rigid")
        output.append("# Group-1: Number of atoms")
        output.append(natoms)
        output.append("# Group-1: Atomic positions")
        for i, line in enumerate(ff_dict["atomic_positions"]):
            output.append(" ".join([str(x) for x in [i] + line]))
        output.append(
            "# Chiral centers Bond  BondDipoles Bend  UrayBradley InvBend  Torsion Imp. Torsion Bond/Bond "
            + "Stretch/Bend Bend/Bend Stretch/Torsion Bend/Torsion IntraVDW IntraCoulomb"
        )
        output.append(
            " ".join([str(x) for x in [0] + [natoms - 1] + [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]])
        )
        if natoms > 1:
            output.append("# Bond stretch: atom n1-n2, type, parameters")
            for i in range(1, natoms):
                output.append("0 {} RIGID_BOND".format(i))
        output.append("# Number of config moves")
        output.append(0)
    string = "\n".join([str(x) for x in output]) + "\n"
    return string


def ff_builder(params: dict) -> dict:
    """Build the force field files.

    Args:
        params (dict): Input parameters, for example:
            params = {
                # See force fields available in ff_data.yaml as framework.keys()
                'ff_framework': 'UFF',
                # See molecules available in ff_data.yaml as ff_data.keys()
                'ff_molecules': {
                    'CO2': 'TraPPE',
                # See force fields available in ff_data.yaml as {molecule}.keys()
                    'N2': 'TraPPE',
                },
                # If True shift despersion interactions, if False simply truncate them.
                'shifted': True,                    .
                # If True apply tail corrections based on homogeneous-liquid assumption
                'tail_corrections': False,
                 Options: 'Lorentz-Berthelot' or 'Jorgensen'
                'mixing_rule': 'Lorentz-Berthelot'
                 # If True use framework's force field for framework-molecule interactions
                'separate_interactions': True
            }

    Returns:
        dict: ff settings
    """
    ff_data = load_yaml()

    out_dict = {}
    out_dict["force_field_mixing_rules_def"], ff_mix_found = render_ff_mixing_def(ff_data, params)
    out_dict["force_field_def"] = render_ff_def(ff_data, params, ff_mix_found)
    out_dict["pseudo_atoms_def"] = render_pseudo_atoms_def(ff_data, params)
    for molecule_name in params["ff_molecules"]:
        out_dict["molecule_{}_def".format(molecule_name)] = render_molecule_def(
            ff_data, params, molecule_name
        )

    return out_dict



# -*- coding: utf-8 -*-
"""Featurizers describing the chemistry of a MOF"""
from .amd import AMD  # noqa: F401
from .aprdf import APRDF  # noqa: F401
from .energygrid import EnergyGridHistogram  # noqa: F401
from .henry import Henry  # noqa: F401
from .partialchargehistogram import PartialChargeHistogram  # noqa: F401
from .partialchargestats import PartialChargeStats  # noqa: F401
from .price import PriceLowerBound  # noqa: F401
from .racs import RACS  # noqa: F401



# -*- coding: utf-8 -*-
"""Featurizers operating on the framework and guests."""
from .guest_aprdf import GuestCenteredAPRDF  # noqa: F401
from .host_guest_featurizer import HostGuestFeaturizer  # noqa: F401



# -*- coding: utf-8 -*-
"""Test that the nconf20 featurizer works on molecules."""

from rdkit.Chem import AllChem

from mofdscribe.featurizers.bu.nconf20_featurizer import NConf20, _n_conf20


def test_nconf20_featurizer(linker_molecule):
    """Test that the nconf20 featurizer works on molecules."""
    # first test that it works directly with RDKit molecules
    # structure (latanoprost) is from Fig 1 in https://www.nature.com/articles/s41597-022-01288-4
    mol = AllChem.MolFromSmiles(
        r"CC(C)OC(=O)CCC/C=C\C[C@H]1[C@H](C[C@H]([C@@H]1CC[C@H](CCC2=CC=CC=C2)O)O)O"
    )
    nconf20 = _n_conf20(mol)
    assert nconf20[0] > 50

    featurizer = NConf20()
    feats = featurizer.featurize(linker_molecule)
    assert len(feats) == 1
    assert feats[0] == 0

    assert len(featurizer.citations()) == 3



# -*- coding: utf-8 -*-

from pymatgen.core import Element

from mofdscribe.featurizers.chemistry._fragment import (
    get_bb_indices,
    get_bbs_from_indices,
    get_floating_indices,
    get_node_atoms,
)
from mofdscribe.featurizers.utils.structure_graph import get_structure_graph


def test_get_floating_indices(hkust_structure, irmof_structure, abacuf_structure):
    for structure in [hkust_structure, irmof_structure, abacuf_structure]:
        sg = get_structure_graph(structure, "vesta")

        assert len(get_floating_indices(sg)) == 0


def test_get_node_atoms(hkust_structure):
    sg = get_structure_graph(hkust_structure, "vesta")
    node_atoms = get_node_atoms(sg)
    assert len(node_atoms) < len(sg)
    assert isinstance(node_atoms, set)
    for site in node_atoms:
        assert str(hkust_structure[site].specie) == "Cu"
    assert len(node_atoms) == dict(hkust_structure.composition)[Element("Cu")]


def test_get_bbs_from_indices(hkust_structure):
    sg = get_structure_graph(hkust_structure, "vesta")
    node_atoms = get_node_atoms(sg)
    all_atoms = set(list(range(len(sg))))
    linker_atoms = all_atoms - node_atoms
    assert linker_atoms & node_atoms == set()
    (
        linker_mol,
        linker_subgraph,
        linker_idx,
        linker_center,
        linker_coordinate,
    ) = get_bbs_from_indices(sg, linker_atoms)
    flat_linker_idx = sum(linker_idx, [])
    assert set(flat_linker_idx) & set(node_atoms) == set()
    assert len(sg) == len(node_atoms) + len(set(flat_linker_idx))
    for idx in flat_linker_idx:
        assert str(hkust_structure[idx].specie) in ("H", "C", "O", "N")


def test_get_bb_indices(hkust_structure):
    sg = get_structure_graph(hkust_structure, "vesta")
    bb_indices = get_bb_indices(sg)
    assert len(bb_indices["nodes"]) < len(sg)
    assert len(bb_indices["nodes"]) > 0
    assert (
        len(set(sum(bb_indices["nodes"], []))) == dict(hkust_structure.composition)[Element("Cu")]
    )

    for site in sum(bb_indices["linker_all"], []):
        assert str(hkust_structure[site].specie) in ("H", "C", "O", "N")

    assert (
        len(set(sum(bb_indices["linker_all"], [])))
        == len(sg) - dict(hkust_structure.composition)[Element("Cu")]
    )

    assert (
        len(sg) - len(set(sum(bb_indices["nodes"], []))) - len(sum(bb_indices["linker_all"], []))
        == 0
    )

    assert len(sum(bb_indices["linker_functional"], [])) == 0
    assert len(sum(bb_indices["linker_all"], [])) == len(
        sum(bb_indices["linker_scaffold"], [])
    ) + len(sum(bb_indices["linker_connecting"], []))



# -*- coding: utf-8 -*-
"""Structures from the ARABG database and labels from Moosavi et al."""
import os
from typing import Collection, Optional, Tuple

import numpy as np
import pandas as pd
from loguru import logger

from mofdscribe.constants import MOFDSCRIBE_PYSTOW_MODULE
from mofdscribe.datasets.checks import check_all_file_exists, length_check
from mofdscribe.datasets.dataset import AbstractStructureDataset
from mofdscribe.datasets.utils import compress_dataset

__all__ = ["ARABGDataset"]


class ARABGDataset(AbstractStructureDataset):
    """
    Exposes the ARABG dataset used in [Moosavi2020]_.

    The raw labels and structures can accessed also on
    `MaterialsCloud <https://archive.materialscloud.org/record/2020.67>`_.

    It is a subset of the ARABG database [Anderson2018]_ with labels
    computed by Moosavi et al.
    Those labels deviate in value and computational
    approach from the original labels in [Boyd2019]_ but are consistent
    with the labels for the other databases in [Moosavi2020]_.

    The available labels are:
        * 'pure_CO2_kH': Henry coefficient of CO2 obtained by Widom method in mol kg-1 Pa-1
        * 'pure_CO2_widomHOA': Heat of adsorption of CO2 obtained by Widom method in
        * 'pure_methane_kH': Henry coefficient of methane obtained by Widom method in mol kg-1 Pa-1
        * 'pure_methane_widomHOA': Heat of adsorption of methane obtained by Widom method
        * 'pure_uptake_CO2_298.00_15000': Pure CO2 uptake at 298.00 K and 15000 Pa in mol kg-1
        * 'pure_uptake_CO2_298.00_1600000': Pure CO2 uptake at 298.00 K and 1600000 Pa in mol kg-1
        * 'pure_uptake_methane_298.00_580000': Pure methane uptake at 298.00 K and 580000 Pa in mol kg-1
        * 'pure_uptake_methane_298.00_6500000': Pure methane uptake at 298.00 K and 6500000 Pa in mol kg-1
        * 'logKH_CO2': Logarithm of Henry coefficient of CO2 obtained by Widom method
        * 'logKH_CH4': Logarithm of Henry coefficient of methane obtained by Widom method
        * 'CH4DC': CH4 deliverable capacity in vSTP/v
        * 'CH4HPSTP': CH4 high pressure uptake in standard temperature and pressure in vSTP/v
        * 'CH4LPSTP': CH4 low pressure uptake in standard temperature and pressure in vSTP/v

    .. note::

        The ARABG structures are hypothetical MOFs, therefore the following caveats apply:

        * It is well known that the data distribution can be quite different from experimental
          structures
        * The structures were only optimized using the UFF force field [UFF]_
        * A time-based split cannot be used for hypothetical structures


    .. admonition:: Information about building blocks
        :class: hint

        This dataset exposed information about the building blocks of the MOFs.
        You might find this useful for grouped-cross-validation (as MOFs with same building-blocks
        and/or net are not really independent).

        You find this info under also in the `info.rcsr_code`, `info.metal_bb`, and
        `info.organic_bb`, `info.functional_group` columns.

    .. warning:: Danger of data leakage

        Cross validation for MOFs with same building-blocks and/or net is notoriously
        difficult. Since all combinations of building-blocks and/or net are considered,
        it is not trivial to find completely independent groups.

    References:
        .. [Moosavi2020] `Moosavi, S. M.; Nandy, A.; Jablonka, K. M.; Ongari, D.; Janet, J. P.; Boyd, P. G.; Lee,
            Y.; Smit, B.; Kulik, H. J. Understanding the Diversity of the Metal-Organic Framework Ecosystem.
            Nature Communications 2020, 11 (1), 4068. <https://doi.org/10.1038/s41467-020-17755-8>`_

        .. [Anderson2018] `Anderson, R.; Rodgers, J.; Argueta, E.; Biong, A.; Gómez-Gualdrón, D. A.
            Role of Pore Chemistry and Topology in the CO2 Capture Capabilities of MOFs:
            From Molecular Simulation to Machine Learning. Chemistry of Materials 2018, 30 (18), 6325–6337.
            <https://doi.org/10.1021/acs.chemmater.8b02257>_
    """

    _files = {
        "v0.0.1": {
            "df": "https://zenodo.org/record/7032106/files/data.json?download=1",
            "structures": "https://zenodo.org/record/7032106/files/structures.tar.gz?download=1",
            "expected_length": 370,
        }
    }

    def __init__(
        self,
        version: str = "v0.0.1",
        drop_graph_duplicates: bool = True,
        subset: Optional[Collection[int]] = None,
    ):
        """Construct an instance of the CoRE dataset.

        Args:
            version (str): version number to use.
                Defaults to "v0.0.1".
            drop_graph_duplicates (bool): If True, keep only one structure
                per decorated graph hash. Defaults to True.
            subset (Collection[int], optional): indices of the structures to include.
                Defaults to None.

        Raises:
            ValueError: If the provided version number is not available.
        """
        self._drop_graph_duplicates = drop_graph_duplicates
        if version not in self._files:
            raise ValueError(
                f"Version {version} not available. Available versions: {list(self._files.keys())}"
            )
        self.version = version

        self._structure_dir = MOFDSCRIBE_PYSTOW_MODULE.ensure_untar(
            "ARABG",
            self.version,
            name="structures.tar.gz",
            url=self._files[version]["structures"],
        )

        self._df = pd.DataFrame(
            MOFDSCRIBE_PYSTOW_MODULE.ensure_json(
                "ARABG", self.version, name="data.json", url=self._files[version]["df"]
            )
        ).reset_index(drop=True)
        compress_dataset(self._df)

        length_check(self._df, self._files[version]["expected_length"])

        if drop_graph_duplicates:
            old_len = len(self._df)
            self._df = self._df.drop_duplicates(subset=["info.decorated_graph_hash"])
            logger.debug(
                f"Dropped {old_len - len(self._df)} duplicate graphs. New length {len(self._df)}"
            )
        self._df = self._df.reset_index(drop=True)
        if subset is not None:
            self._df = self._df.iloc[subset]
            self._df = self._df.reset_index(drop=True)

        self._structures = [
            os.path.join(self._structure_dir, f + ".cif") for f in self._df["info.name"]
        ]

        check_all_file_exists(self._structures)

        self._decorated_graph_hashes = self._df["info.decorated_graph_hash"].values
        self._undecorated_graph_hashes = self._df["info.undecorated_graph_hash"].values
        self._decorated_scaffold_hashes = self._df["info.decorated_scaffold_hash"].values
        self._undecorated_scaffold_hashes = self._df["info.undecorated_scaffold_hash"].values
        self._densities = self._df["info.density"].values
        self._labelnames = (c for c in self._df.columns if c.startswith("outputs."))
        self._featurenames = (c for c in self._df.columns if c.startswith("features."))
        self._infonames = (c for c in self._df.columns if c.startswith("info."))

    def get_subset(self, indices: Collection[int]) -> "AbstractStructureDataset":
        """Get a subset of the dataset.

        Args:
            indices (Collection[int]): indices of the structures to include.

        Returns:
            AbstractStructureDataset: a new dataset containing only the structures
                specified by the indices.
        """
        return ARABGDataset(
            version=self.version,
            drop_graph_duplicates=self._drop_graph_duplicates,
            subset=indices,
        )

    @property
    def available_info(self) -> Tuple[str]:
        return self._infonames

    @property
    def available_features(self) -> Tuple[str]:
        return self._featurenames

    @property
    def available_labels(self) -> Tuple[str]:
        return self._labelnames

    def get_labels(self, idx: Collection[int], labelnames: Collection[str] = None) -> np.ndarray:
        labelnames = labelnames if labelnames is not None else self._labelnames
        return self._df.iloc[idx][list(labelnames)].values

    @property
    def citations(self) -> Tuple[str]:
        return [
            "@article{Moosavi2020,"
            "doi = {10.1038/s41467-020-17755-8},"
            "url = {https://doi.org/10.1038/s41467-020-17755-8},"
            "year = {2020},"
            "month = aug,"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {11},"
            "number = {1},"
            "author = {Seyed Mohamad Moosavi and Aditya Nandy and Kevin Maik Jablonka "
            "and Daniele Ongari and Jon Paul Janet and Peter G. Boyd and Yongjin Lee "
            "and Berend Smit and Heather J. Kulik},"
            "title = {Understanding the diversity of the metal-organic framework ecosystem},"
            "journal = {Nature Communications}"
            "}",
            "@article{Anderson_2018,"
            "doi = {10.1021/acs.chemmater.8b02257},"
            "url = {https://doi.org/10.1021%2Facs.chemmater.8b02257},"
            "year = 2018,"
            "month = {aug},"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {30},"
            "number = {18},"
            "pages = {6325--6337},"
            "author = {Ryther Anderson and Jacob Rodgers and Edwin Argueta "
            "and Achay Biong and Diego A. G{'{o}}mez-Gualdr{'{o}}n},"
            "title = {Role of Pore Chemistry and Topology in the {CO}$_2$ "
            "Capture Capabilities of {MOFs}: From Molecular Simulation to Machine Learning},"
            "journal = {Chem. Mater.}"
            "}",
        ]



"""Base featurizer for MOF structure based featurizers.

The main purpose of these classes is currently that they handle
conversion of the structures to primtive cells.
This can have computational benefits and also make the use of some
aggregations such as "sum" more meaningful.

However, since they support this functionality, they are less flexible
than the "original" matminer :code:`BaseFeaturizer` and :code:`MultipleFeaturizer`.
In practice, this means that they only accept one pymatgen structure object
or and iterable of pymatgen structure objects.
"""
from abc import abstractmethod
from functools import partial
from multiprocessing import Pool
from typing import Union

import numpy as np
import pandas as pd
from loguru import logger
from matminer.featurizers.base import BaseFeaturizer, MultipleFeaturizer
from pymatgen.core import IMolecule, IStructure, Molecule, Structure
from tqdm.auto import tqdm


def get_primitive(structure: Union[IStructure, Structure, Molecule, IMolecule]) -> Structure:
    """Get the primitive cell of a structure.

    We use this wrapper because we want to have a passtrough
    for molecules, which some of our featurizers also accept.

    Args:
        structure: Structure object.

    Returns:
        Structure object.
    """
    if isinstance(structure, (Structure, IStructure)):
        return structure.get_primitive_structure()
    else:
        return structure


class MOFBaseFeaturizer(BaseFeaturizer):
    """Base featurizer for MOF structure based featurizers.

    .. note::

        If you implement a new :code:`MOFBaseFeaturizer`,
        you need to implement a :code:`_featurize` method.
        If you implement :code:`featurize` directly,
        you would override the conversion to primitive cell.

    .. warning::

        If you implement a new :code:`MOFBaseFeaturizer`,
        and your featurizer needs to be fitted, keep in mind
        to also call the :code:`_get_primitive` method.
    """

    def __init__(self, primitive: bool = False) -> None:
        """
        Construct a MOFBaseFeaturizer.

        Args:
            primitive (bool): If True, use the primitive cell of the structure.
                Defaults to False.
        """
        self.primitive = primitive

    def _get_primitive_many(self, structures):
        if self.n_jobs == 1:
            return [self._get_primitive(s) for s in structures]
        else:
            with Pool(self.n_jobs, maxtasksperchild=1) as p:
                res = p.map(self._get_primitive, structures, chunksize=self.chunksize)
                return res

    def _get_primitive(self, structure: Union[Structure, IStructure]) -> Structure:
        logger.debug("Getting primitive cell for structure in MOFBaseFeaturizer")
        return get_primitive(structure)

    @abstractmethod
    def _featurize(self, structure: Union[Structure, IStructure]) -> np.ndarray:
        raise NotImplementedError("_featurize must be implemented in a subclass")

    def fit(self, structures):
        if not isinstance(structures, (list, tuple)):
            structures = [structures]
        if self.primitive:
            structures = self._get_primitive_many(structures)
        self._fit(structures)

    def featurize(self, structure: Union[Structure, IStructure]) -> np.ndarray:
        """Compute the descriptor for a given structure.

        Args:
            structure (Union[Structure, IStructure]): Structure to compute the descriptor for.

        Returns:
            A numpy array containing the descriptor.
        """
        logger.debug("Featurizing structure in MOFBaseFeaturizer")
        if self.primitive:
            logger.debug("Getting primitive cell for structure in MOFBaseFeaturizer")
            structure = get_primitive(structure)
        return self._featurize(structure)

    def featurize_many(self, entries, ignore_errors=False, return_errors=False, pbar=True):
        """Featurize a list of entries.

        If `featurize` takes multiple inputs, supply inputs as a list of tuples.

        Featurize_many supports entries as a list, tuple, numpy array,
        Pandas Series, or Pandas DataFrame.

        Args:
            entries (list-like object): A list of entries to be featurized.
            ignore_errors (bool): Returns NaN for entries where exceptions are
                thrown if True. If False, exceptions are thrown as normal.
            return_errors (bool): If True, returns the feature list as
                determined by ignore_errors with traceback strings added
                as an extra 'feature'. Entries which featurize without
                exceptions have this extra feature set to NaN.
            pbar (bool): Show a progress bar for featurization if True.

        Returns:
            (list) features for each entry.

        Raises:
            Exception: If entries is not a list-like object.
            ValueError: If return_errors is set and ignore_errors is True.
        """
        if return_errors and not ignore_errors:
            raise ValueError("Please set ignore_errors to True to use" " return_errors.")

        # Check inputs
        if not isinstance(entries, (tuple, list, np.ndarray, pd.Series, pd.DataFrame)):
            raise Exception("'entries' must be a list-like object")

        # Special case: Empty list
        if len(entries) == 0:
            return []

        # If the featurize function only has a single arg, zip the inputs
        if isinstance(entries, pd.DataFrame):
            entries = entries.values
        elif isinstance(entries, pd.Series) or not isinstance(
            entries[0], (tuple, list, np.ndarray)
        ):
            entries = zip(entries)

        # Add a progress bar
        if pbar:
            # list() required, tqdm has issues with memory if generator given
            entries = tqdm(list(entries), desc=self.__class__.__name__)

        # Run the actual featurization
        if self.n_jobs == 1:
            return np.array(
                [
                    self.featurize_wrapper(
                        x, ignore_errors=ignore_errors, return_errors=return_errors
                    )
                    for x in entries
                ]
            )
        else:
            with Pool(self.n_jobs, maxtasksperchild=1) as p:
                func = partial(
                    self.featurize_wrapper,
                    return_errors=return_errors,
                    ignore_errors=ignore_errors,
                )
                res = p.map(func, entries, chunksize=self.chunksize)
                return np.array(res)


class MOFMultipleFeaturizer(MultipleFeaturizer):
    """Base multiplefeaturizer for MOF structure based featurizers.

    .. warning::

        This MultipleFeaturizer only works with featurizers that only
        accept single structures as input.

    .. warning::

        This MultipleFeaturizer cannot efficiently be used if you want
        to use the primitive for some featurizers and not for others.
        In this case, you're better off using two separate MultipleFeaturizers.
    """

    def __init__(self, featurizers, iterate_over_entries=True, primitive=True):
        """
        Construct a MOFMultipleFeaturizer.

        Args:
            featurizers (list): List of featurizers to use.
            iterate_over_entries (bool): If True, featurize each entry in the list.
                If False, featurize each structure in the list. Defaults to True.
            primitive (bool): If True, use the primitive cell of the structure.
                Defaults to True.
        """
        # unset the primitive on the individual featurizers
        for featurizer in featurizers:
            featurizer.primitive = False

        self.featurizers = featurizers
        self.iterate_over_entries = iterate_over_entries
        self.primitive_multiple = primitive

    def _get_primitive(self, structure: Union[Structure, IStructure]) -> Structure:
        return get_primitive(structure)

    def featurize(self, structure: Union[Structure, IStructure]) -> np.ndarray:
        logger.debug(f"Featurizing structure in MOFMultipleFeaturizer, {self.primitive_multiple}")
        if self.primitive_multiple:
            logger.debug("Getting primitive cell")
            structure = self._get_primitive(structure)
        return np.array([feature for f in self.featurizers for feature in f.featurize(structure)])

    def _get_primitive_many(self, structures):
        if self.n_jobs == 1:
            return [self._get_primitive(s) for s in structures]
        else:
            with Pool(self.n_jobs, maxtasksperchild=1) as p:
                res = p.map(self._get_primitive, structures, chunksize=self.chunksize)
                return res

    def featurize_many(self, entries, ignore_errors=False, return_errors=False, pbar=True):
        logger.debug("Precomputing primitive cells")
        if self.primitive_multiple:
            entries = self._get_primitive_many(entries)
        if self.iterate_over_entries:
            return np.array(
                super().featurize_many(
                    entries,
                    ignore_errors=ignore_errors,
                    return_errors=return_errors,
                    pbar=pbar,
                )
            )
        else:
            features = [
                f.featurize_many(
                    entries,
                    ignore_errors=ignore_errors,
                    return_errors=return_errors,
                    pbar=pbar,
                )
                for f in self.featurizers
            ]

            return np.hstack(features)

    def featurize_wrapper(self, x, return_errors=False, ignore_errors=False):
        if self.iterate_over_entries:
            return np.array(
                [
                    feature
                    for f in self.featurizers
                    for feature in f.featurize_wrapper(
                        x, return_errors=return_errors, ignore_errors=ignore_errors
                    )
                ]
            )
        else:
            return super().featurize_wrapper(
                x, return_errors=return_errors, ignore_errors=ignore_errors
            )



# -*- coding: utf-8 -*-
"""Test the helpers for running simulations with RASPA"""
from mofdscribe.featurizers.utils.raspa.resize_uc import resize_unit_cell


def test_resize_unit_cell(hkust_structure):
    """Test the resize_unit_cell function"""
    supercell = resize_unit_cell(hkust_structure, 28)
    assert supercell == (2, 2, 2)



# -*- coding: utf-8 -*-
from .geometric_properties import AccessibleVolume  # noqa: F401
from .geometric_properties import PoreDiameters  # noqa: F401
from .geometric_properties import PoreSizeDistribution  # noqa: F401
from .geometric_properties import RayTracingHistogram  # noqa: F401
from .geometric_properties import SurfaceArea  # noqa: F401



# -*- coding: utf-8 -*-
"""Utlities for working with persistence diagrams."""
from collections import defaultdict
from typing import Collection, Dict, List, Optional, Tuple

import numpy as np
from element_coder import encode_many
from loguru import logger
from pymatgen.core import Structure
from pymatgen.transformations.advanced_transformations import CubicSupercellTransformation

from mofdscribe.featurizers.utils.aggregators import MA_ARRAY_AGGREGATORS
from mofdscribe.featurizers.utils.substructures import filter_element


# @np_cache
def construct_pds_cached(coords, periodic=False, weights: Optional[Collection] = None):
    from moleculetda.construct_pd import construct_pds

    return construct_pds(coords, periodic=periodic, weights=weights)


def _coords_for_structure(
    structure: Structure,
    min_size: int = 50,
    periodic: bool = False,
    no_supercell: bool = False,
    weighting: Optional[str] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    from moleculetda.read_file import make_supercell

    if no_supercell:
        if weighting is not None:
            weighting = encode_many([str(s.symbol) for s in structure.species], weighting)
        return structure.cart_coords, weighting

    else:
        if periodic:
            transformed_s = CubicSupercellTransformation(min_size=min_size).apply_transformation(
                structure
            )
            if weighting is not None:
                weighting = encode_many([str(s.symbol) for s in transformed_s.species], weighting)
            return transformed_s.cart_coords, weighting
        else:
            if weighting is not None:
                weighting_arr = np.array(
                    encode_many([str(s.symbol) for s in structure.species], weighting)
                )
                # we can add the weighing as additional column for the cooords
                coords_w_weight = make_supercell(
                    np.hstack([structure.cart_coords, weighting_arr.reshape(-1, 1)]),
                    structure.lattice.matrix,
                    min_size,
                )
                return coords_w_weight[:, :-1], coords_w_weight[:, -1]
            else:
                return (
                    make_supercell(structure.cart_coords, structure.lattice.matrix, min_size),
                    None,
                )


def _pd_arrays_from_coords(
    coords, periodic: bool = False, bd_arrays: bool = False, weights: Optional[np.ndarray] = None
):
    from moleculetda.vectorize_pds import diagrams_to_arrays

    pds = construct_pds_cached(coords, periodic=periodic, weights=weights)
    if bd_arrays:
        pd = diagrams_to_bd_arrays(pds)
    else:
        pd = diagrams_to_arrays(pds)

    return pd


# ToDo: only do this for selected elements
# ToDo: only do this for all if we want
def get_persistent_images_for_structure(
    structure: Structure,
    elements: List[List[str]],
    compute_for_all_elements: bool = True,
    min_size: int = 20,
    spread: float = 0.2,
    weighting: str = "identity",
    pixels: Tuple[int] = (50, 50),
    max_b: int = 18,
    max_p: int = 18,
    periodic: bool = False,
    no_supercell: bool = False,
    alpha_weighting: Optional[str] = None,
) -> dict:
    """
    Get the persistent images for a structure.

    Args:
        structure (Structure): input structure
        elements (List[List[str]]): list of elements to compute for
        compute_for_all_elements (bool): compute for all elements
        min_size (int): minimum size of the cell for construction of persistent images
        spread (float): spread of kernel for construction
            of persistent images
        weighting (str): weighting scheme for construction
            of persistent images
        pixels (Tuple[int]): size of the image in pixels
        max_b (int): maximum birth time for construction of persistent images
        max_p (int): maximum persistence time for construction of persistent
            images
        periodic (bool): if True (experimental!), use the periodic
            Euclidean distance
        no_supercell (bool): if True, then supercell expansion is not performed.
            The preceeding min_size argument is then ignored.
            Defaults to False.
        alpha_weighting (str, optional): if given use weighted alpha shapes,
            e.g., `atomic_radius_calculated` or `van_der_waals_radius`.
            Defaults to None.

    Returns:
        persistent_images (dict): dictionary of persistent images and their
            barcode representations
    """
    from moleculetda.vectorize_pds import get_images

    element_images: Dict[dict] = defaultdict(dict)
    specs = []
    for mb, mp in zip(max_b, max_p):
        specs.append({"minBD": 0, "maxB": mb, "maxP": mp})
    for element in elements:
        try:
            filtered_structure = filter_element(structure, element)
            coords, weights = _coords_for_structure(
                filtered_structure,
                min_size=min_size,
                periodic=periodic,
                no_supercell=no_supercell,
                weighting=alpha_weighting,
            )
            pd = _pd_arrays_from_coords(coords, periodic=periodic)

            images = get_images(
                pd,
                spread=spread,
                weighting=weighting,
                pixels=pixels,
                specs=specs,
            )
        except ValueError:
            logger.exception(f"Error computing persistent images for {element}")
            images = np.zeros((0, pixels[0], pixels[1]))
            images[:] = np.nan
            pd = np.zeros((0, max_p + 1))
            pd[:] = np.nan

        # ToDo: make sure that we have the correct length
        element_images["image"][element] = images
        element_images["array"][element] = pd

    if compute_for_all_elements:
        coords, weights = _coords_for_structure(
            structure,
            min_size=min_size,
            periodic=periodic,
            no_supercell=no_supercell,
            weighting=alpha_weighting,
        )
        pd = _pd_arrays_from_coords(coords, periodic=periodic)

        images = get_images(pd, spread=spread, weighting=weighting, pixels=pixels, specs=specs)
        element_images["image"]["all"] = images
        element_images["array"]["all"] = pd

    return element_images


def get_min_max_from_dia(dia, birth_persistence: bool = True):
    if len(dia) == 0:
        return [0, 0, 0, 0]
    d = np.array([[x["birth"], x["death"]] for x in dia])

    if birth_persistence:
        # convert to birth - persistence
        d[:, 1] -= d[:, 0]
    d = np.ma.masked_invalid(d)
    return [d[:, 0].min(), d[:, 0].max(), d[:, 1].min(), d[:, 1].max()]


def diagrams_to_bd_arrays(dgms):
    """Convert persistence diagram objects to persistence diagram arrays."""
    dgm_arrays = {}
    for dim, dgm in enumerate(dgms):

        if dgm:
            arr = np.array(
                [[np.sqrt(dgm[i].birth), np.sqrt(dgm[i].death)] for i in range(len(dgm))]
            )

            mask = np.isfinite(arr).all(axis=1)

            arr = arr[mask]
            dgm_arrays[f"dim{dim}"] = arr

        else:
            dgm_arrays[f"dim{dim}"] = np.zeros((0, 2))

    return dgm_arrays


def get_diagrams_for_structure(
    structure,
    elements: List[List[str]],
    compute_for_all_elements: bool = True,
    min_size: int = 20,
    periodic: bool = False,
    no_supercell: bool = False,
    alpha_weighting: Optional[str] = None,
):
    keys = [f"dim{i}" for i in range(3)]
    element_dias = defaultdict(dict)
    nan_array = np.zeros((0, 2))
    nan_array[:] = np.nan
    for element in elements:
        try:
            filtered_structure = filter_element(structure, element)
            coords, weights = _coords_for_structure(
                filtered_structure,
                min_size=min_size,
                periodic=periodic,
                no_supercell=no_supercell,
                weighting=alpha_weighting,
            )
            arrays = _pd_arrays_from_coords(
                coords, periodic=periodic, bd_arrays=True, weights=weights
            )
        except Exception:
            logger.exception(f"Error for element {element}")
            arrays = {key: nan_array for key in keys}
        if not len(arrays) == 4:
            for key in keys:
                if key not in arrays:
                    arrays[key] = nan_array
        element_dias[element] = arrays

    if compute_for_all_elements:
        coords, weights = _coords_for_structure(
            structure,
            min_size=min_size,
            periodic=periodic,
            no_supercell=no_supercell,
            weighting=alpha_weighting,
        )
        arrays = _pd_arrays_from_coords(coords, periodic=periodic, bd_arrays=True, weights=weights)
        element_dias["all"] = arrays
        if len(arrays) != 4:
            for key in keys:
                if key not in arrays:
                    arrays[key] = nan_array
    if len(element_dias) != len(elements) + int(compute_for_all_elements):
        raise ValueError("Something went wrong with the diagram extraction.")
    return element_dias


def get_persistence_image_limits_for_structure(
    structure: Structure,
    elements: List[List[str]],
    compute_for_all_elements: bool = True,
    min_size: int = 20,
    periodic: bool = False,
    no_supercell: bool = False,
    alpha_weighting: Optional[str] = None,
) -> dict:
    limits = defaultdict(list)
    for element in elements:
        try:
            filtered_structure = filter_element(structure, element)

            coords, weights = _coords_for_structure(
                filtered_structure,
                min_size=min_size,
                periodic=periodic,
                no_supercell=no_supercell,
                weighting=alpha_weighting,
            )
            pd = _pd_arrays_from_coords(coords, periodic=periodic, weights=weights)
            for k, v in pd.items():
                limits[k].append(get_min_max_from_dia(v))
        except ValueError:
            logger.exception("Could not extract diagrams for element %s", element)
            pass

    if compute_for_all_elements:
        coords, weights = _coords_for_structure(
            structure,
            min_size=min_size,
            periodic=periodic,
            no_supercell=no_supercell,
            weighting=alpha_weighting,
        )
        pd = _pd_arrays_from_coords(coords, periodic=periodic, weights=weights)
        for k, v in pd.items():
            limits[k].append(get_min_max_from_dia(v))
    return limits


def persistent_diagram_stats(
    diagram: np.ndarray, aggregrations: Tuple[str], nanfiller: float = 0
) -> dict:
    """
    Compute statistics for a persistence diagram.

    Args:
        diagram (np.ndarray): The persistence diagram.
        aggregrations (Tuple[str]): The name of the aggregations to compute.
        nanfiller (float): The value to use when a nan is encountered.

    Returns:
        dict: nested dictionary with the following structure:
            {'persistence_parameter': {'statistic': value}}
        where persistence_parameter is one of ['birth', 'death', 'persistence']
    """
    stats = {
        "birth": {},
        "death": {},
        "persistence": {},
    }

    try:
        d = np.array([[x["birth"], x["death"], x["death"] - x["birth"]] for x in diagram])
    except IndexError:
        d = np.array([[x[0], x[1], x[1] - x[0]] for x in diagram])
    d = np.ma.masked_invalid(d)

    for aggregation in aggregrations:
        agg_func = MA_ARRAY_AGGREGATORS[aggregation]
        for i, key in enumerate(["birth", "death", "persistence"]):
            try:
                stats[key][aggregation] = agg_func(d[:, i])
            except IndexError:
                stats[key][aggregation] = nanfiller
    return stats



# -*- coding: utf-8 -*-
"""Test the pairwise distance histogram featurizer."""
from mofdscribe.featurizers.bu.distance_hist_featurizer import PairwiseDistanceHist


def test_pairwise_distance_hist_featurizer(molecule, linker_molecule, triangle_structure):
    """Test the pairwise distance hist featurizer."""
    for mol in (molecule, linker_molecule, triangle_structure):
        featurizer = PairwiseDistanceHist()
        feats = featurizer.featurize(mol)
        assert len(feats) == 30
        for f in feats:
            assert f >= 0
        assert feats.sum() >= 1
        assert len(feats) == len(featurizer.feature_labels())



# -*- coding: utf-8 -*-
"""Test the EnergyGridHistogram featurizer."""
import os

from mofdscribe.featurizers.chemistry.energygrid import EnergyGridHistogram, read_ascii_grid

from ..helpers import is_jsonable

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def test_read_ascii_grid():
    """Ensure that we can parse an ASCII grid file."""
    file_name = os.path.join(THIS_DIR, "..", "..", "test_files", "asci_grid_C_co2.grid")
    result = read_ascii_grid(file_name)
    assert len(result) == 22185
    assert result["energy"].dtype == float


def test_energygrid(hkust_structure):
    """Make sure that the featurization works for typical MOFs and the number of features is as expected."""
    eg = EnergyGridHistogram()
    feats = eg.featurize(hkust_structure)

    assert len(feats) == 40
    assert len(eg.feature_labels()) == 40
    assert is_jsonable(dict(zip(eg.feature_labels(), feats)))
    assert feats.ndim == 1



# -*- coding: utf-8 -*-
"""Test the pairwise distance stats featurizer."""
from mofdscribe.featurizers.bu.distance_stats_featurizer import PairwiseDistanceStats


def test_pairwise_distance_stats_featurizer(molecule, linker_molecule, triangle_structure):
    """Test the pairwise distance stats featurizer."""
    featurizer = PairwiseDistanceStats()
    feats = featurizer.featurize(molecule)
    assert len(feats) == 4
    for f in feats:
        assert f > 0

    feats = featurizer.featurize(linker_molecule)
    assert len(feats) == 4
    for f in feats:
        assert f > 0
    assert len(feats) == len(featurizer.feature_labels())

    feats = featurizer.featurize(triangle_structure)
    assert len(feats) == 4
    for f in feats:
        assert f > 0



# -*- coding: utf-8 -*-
"""Compute features on the host and the guests and then aggregate them."""
from typing import Collection, List, Optional, Tuple, Union

import numpy as np
from loguru import logger
from matminer.featurizers.base import BaseFeaturizer
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.hostguest.utils import HostGuest, _extract_host_guest
from mofdscribe.featurizers.utils import set_operates_on
from mofdscribe.featurizers.utils.aggregators import ARRAY_AGGREGATORS


# ToDo: How do we handle if there is no guest?
# make it optional to remove guests from the structure?
class HostGuestFeaturizer(BaseFeaturizer):
    """
    Compoute features on the host and the guests and then aggregate them.

    This can be useful if you use structures with adsorbed guest molecules as input and want
    to compute adsorption properties such as adsorption energies.

    .. warning::
        Note that we assume that the guest is not covalently bonded to the host.

    .. note::
        If there is only one guest, then there is no need to aggregate the features.
        In this case, only use the "mean" aggregation.

    .. note::


    .. example::
        >>> from mofdscribe.featurizers.hostguest import HostGuestFeaturizer
        >>> from matminer.featurizers.structure.sites import SiteStatsFingerprint
        >>> from pymatgen.core import Structure
        >>> structure = Structure.from_file("tests/data/structures/Co2O3.cif")
        >>> featurizer = HostGuestFeaturizer(featurizer=SiteStatsFingerprint.from_preset("SOAP_formation_energy"),
            aggregations=("mean", "std", "min", "max"))
    """

    def __init__(
        self,
        featurizer: BaseFeaturizer,
        aggregations: Tuple[str] = ("mean", "std", "min", "max"),
        local_env_method: str = "vesta",
        remove_guests: bool = True,
    ) -> None:
        """
        Construct a new BUFeaturizer.

        Args:
            featurizer (BaseFeaturizer): The featurizer to use.
                Currently, we do not support `MultipleFeaturizer`s.
                Please, instead, use multiple BUFeaturizers.
                If you use a featurizer that is not implemented in mofdscribe
                (e.g. a matminer featurizer), you need to wrap using a method
                that describes on which data objects the featurizer can operate on.
                If you do not do this, we default to assuming that it operates on structures.
            aggregations (Tuple[str]): The aggregations to use.
                Must be one of :py:obj:`ARRAY_AGGREGATORS`.
            local_env_method (str): The method to use for the local environment determination
                (to compute the structure graph). Defaults to "vesta".
            remove_guests (bool): Whether to remove the guests from the structure.
                This is useful if you want to compute features on the host only, independent
                of the guests. Defaults to True.
        """
        self._featurizer = featurizer
        self._aggregations = aggregations
        self._local_env_method = local_env_method
        self._remove_guests = remove_guests
        set_operates_on(self, featurizer)

    def feature_labels(self) -> List[str]:
        labels = []
        base_labels = self._featurizer.feature_labels()
        for label in base_labels:
            labels.append(f"host_{label}")
        for bb in ["guest"]:
            for aggregation in self._aggregations:
                for label in base_labels:
                    labels.append(f"{bb}_{aggregation}_{label}")
        return labels

    def _extract_host_guest(
        self,
        structure: Optional[Union[Structure, IStructure]] = None,
        host_guest: Optional[HostGuest] = None,
    ):
        return _extract_host_guest(
            structure=structure,
            host_guest=host_guest,
            remove_guests=self._remove_guests,
            operates_on=self._operates_on,
            local_env_method=self._local_env_method,
        )

    def fit(
        self,
        structures: Collection[Union[Structure, IStructure]],
        host_guests: Optional[Collection[HostGuest]] = None,
    ) -> None:
        """
        Fit the featurizer to the given structures.

        Args:
            structures (Collection[Union[Structure, IStructure]]): The structures to fit to.
            host_guests (Optional[Collection[HostGuest]]): The host_guests to fit to.
                If you provide this, you must not provide structures.
        """
        all_hosts, all_guests = [], []

        if structures is not None:
            for structure in structures:
                host_guest = self._extract_host_guest(structure=structure)
                all_hosts.append(host_guest.host)
                all_guests.extend(host_guest.guests)

        if host_guests is not None:
            for host_guest in host_guests:
                host_guest = self._extract_host_guest(host_guest=host_guest)
                all_hosts.append(host_guest.host)
                all_guests.extend(host_guest.guests)
        self._featurizer.fit(all_hosts + all_guests)

    def featurize(
        self,
        structure: Optional[Union[Structure, IStructure]],
        host_guest: Optional[HostGuest] = None,
    ) -> np.ndarray:
        """
        Compute the features of the host and the guests and aggregate them.

        Args:
            structure (Optional[Union[Structure, IStructure]]): The structure to featurize.
            host_guest (Optional[HostGuest]): The host_guest to featurize.
                If you provide this, you must not provide structure.

        Returns:
            np.ndarray: The features of the host and the guests.

        Raises:
            ValueError: If we cannot detect a host.
        """
        host_guest = self._extract_host_guest(structure=structure, host_guest=host_guest)

        if host_guest.host is None:
            raise ValueError(
                "Did not find a framework. This is required for the host guest featurizer."
            )

        if len(host_guest.guests) == 0:
            logger.warning(
                "Did not find any guests. Make sure that the guest is not covalently bonded to the host."
            )

        guest_feats = []
        host_feats = self._featurizer.featurize(host_guest.host)

        for guest in host_guest.guests:
            guest_feats.append(self._featurizer.featurize(guest))

        aggregated_feats = []
        for aggregation in self._aggregations:
            aggregated_feats.extend(ARRAY_AGGREGATORS[aggregation](guest_feats, axis=0))

        aggregated_feats = np.array(aggregated_feats)
        return np.concatenate((host_feats, aggregated_feats))

    def citations(self) -> List[str]:
        return self._featurizer.citations()

    def implementors(self) -> List[str]:
        return self._featurizer.implementors()



# -*- coding: utf-8 -*-
"""Generalized average minimum distance (AMD) featurizer."""
from typing import List, Tuple, Union

import numpy as np
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer
from mofdscribe.featurizers.utils.extend import operates_on_istructure, operates_on_structure
from mofdscribe.featurizers.utils.substructures import filter_element

__all__ = ["AMD"]


@operates_on_istructure
@operates_on_structure
class AMD(MOFBaseFeaturizer):
    """Implements the generalized average minimum distance (AMD) isometry invariant [Widdowson2022]_.

    Note that it currently does not implement averages according to
    multiplicity of sites (as the original code supports).
    The generalization is to other aggregations of the PDD.

    The AMD is the average of the point-wise distance distribution (PDD) of a
    crystal. The PDD lists distances to neighbouring atoms in order, closest
    first. Hence, the kth AMD value corresponds to the average distance to the
    kth nearest neighbour.

    The descriptors can be computed over the full structure or substructures of
    certain atom types.
    """

    def __init__(
        self,
        k: int = 100,
        atom_types: Tuple[str] = (
            "C-H-N-O",
            "F-Cl-Br-I",
            "Cu-Mn-Ni-Mo-Fe-Pt-Zn-Ca-Er-Au-Cd-Co-Gd-Na-Sm-Eu-Tb-V"
            "-Ag-Nd-U-Ba-Ce-K-Ga-Cr-Al-Li-Sc-Ru-In-Mg-Zr-Dy-W-Yb-Y-"
            "Ho-Re-Be-Rb-La-Sn-Cs-Pb-Pr-Bi-Tm-Sr-Ti-Hf-Ir-Nb-Pd-Hg-"
            "Th-Np-Lu-Rh-Pu",
        ),
        compute_for_all_elements: bool = True,
        aggregations: Tuple[str] = ("mean",),
        primitive: bool = True,
    ) -> None:
        """Initialize the AMD descriptor.

        Args:
            k (int): controls the number of nearest neighbour atoms considered
                for each atom in the unit cell. Defaults to 100.
            atom_types (tuple): Atoms that are used to create substructures
                for which the AMD descriptor is computed.
                Defaults to ( 'C-H-N-O', 'F-Cl-Br-I',
                'Cu-Mn-Ni-Mo-Fe-Pt-Zn-Ca-Er-Au-Cd-Co-Gd-Na-Sm-Eu-Tb-V-Ag-Nd-U-Ba-Ce-K-Ga-
                Cr-Al-Li-Sc-Ru-In-Mg-Zr-Dy-W-Yb-Y-Ho-Re-Be-Rb-La-Sn-Cs-Pb-Pr-Bi-Tm-Sr-Ti-Hf-
                Ir-Nb-Pd-Hg-Th-Np-Lu-Rh-Pu', ).
            compute_for_all_elements (bool): If True, compute the AMD descriptor for
                the original structure with all elements. Defaults to True.
            aggregations (tuple): Aggregations of the AMD descriptor.
                The 'mean' is equivalent to the original AMD. Defaults to ('mean',).
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.
        """
        self.k = k
        atom_types = [] if atom_types is None else atom_types
        self.elements = atom_types
        self.atom_types = (
            list(atom_types) + ["all"] if compute_for_all_elements else list(atom_types)
        )
        self.compute_for_all_elements = compute_for_all_elements
        self.aggregations = aggregations
        super().__init__(primitive=primitive)

    def _get_feature_labels(self) -> List[str]:
        labels = []
        for atom_type in self.atom_types:
            for agg in self.aggregations:
                for i in range(self.k):
                    labels.append(f"amd_{atom_type}_{agg}_{i}")

        return labels

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def _featurize(self, structure: Union[Structure, IStructure]) -> np.ndarray:
        """Compute the AMD descriptor for a given structure.

        Args:
            structure (Union[Structure, IStructure]): Structure to compute the descriptor for.

        Returns:
            A numpy array containing the AMD descriptor.
        """
        from amd._nns import nearest_neighbours
        from amd.calculate import _extract_motif_cell
        from amd.periodicset import PeriodicSet

        def get_pdd(structure, k):
            motif, cell, asymmetric_unit, multiplicities = _extract_motif_cell(
                PeriodicSet(structure.cart_coords, structure.lattice.matrix)
            )
            pdd, _, _ = nearest_neighbours(motif, cell, asymmetric_unit, k)
            return pdd

        # ToDo: we can potentially parallelize this
        all_desc = []
        if len(self.elements) > 0:
            for element in self.elements:
                filtered_structure = filter_element(structure, element)
                if filtered_structure is not None:
                    pdd = get_pdd(filtered_structure, self.k)
                else:
                    pdd = np.empty((1, self.k))
                    pdd[:] = np.nan
                for agg in self.aggregations:
                    all_desc.append(getattr(np, agg)(pdd, axis=0))

        if self.compute_for_all_elements:
            pdd = get_pdd(structure, self.k)
            for agg in self.aggregations:
                all_desc.append(getattr(np, agg)(pdd, axis=0))

        return np.concatenate(all_desc)

    def citations(self):
        return [
            "@article{amd2022,"
            "title = {Average Minimum Distances of periodic point sets - "
            "foundational invariants for mapping periodic crystals},"
            "author = {Daniel Widdowson and Marco M Mosca and Angeles Pulido "
            "and Vitaliy Kurlin and Andrew I Cooper},"
            "journal = {MATCH Communications in Mathematical and in Computer Chemistry},"
            "doi = {10.46793/match.87-3.529W},"
            "volume = {87},"
            "number = {3},"
            "pages = {529-559},"
            "year = {2022}"
            "}"
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka", "Daniel Widdowson"]



# -*- coding: utf-8 -*-
"""
Cached graphs and compression.

Compression code taken from https://www.kaggle.com/code/nickycan/compress-70-of-dataset.
"""

from functools import lru_cache

import numpy as np
import pandas as pd
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.core import IStructure
from structuregraph_helpers.create import get_structure_graph
from structuregraph_helpers.hash import (
    decorated_graph_hash,
    decorated_scaffold_hash,
    undecorated_graph_hash,
    undecorated_scaffold_hash,
)


@lru_cache(maxsize=None)
def get_structure_graph_cached(structure: IStructure) -> StructureGraph:
    return get_structure_graph(structure)


@lru_cache(maxsize=None)
def get_decorated_graph_hash_cached(structure: IStructure) -> str:
    return decorated_graph_hash(get_structure_graph_cached(structure))


@lru_cache(maxsize=None)
def get_undecorated_graph_hash_cached(structure: IStructure) -> str:
    return undecorated_graph_hash(get_structure_graph_cached(structure))


@lru_cache(maxsize=None)
def get_decorated_scaffold_hash_cached(structure: IStructure) -> str:
    return decorated_scaffold_hash(get_structure_graph_cached(structure))


@lru_cache(maxsize=None)
def get_undecorated_scaffold_hash_cached(structure: IStructure) -> str:
    return undecorated_scaffold_hash(get_structure_graph_cached(structure))


_INT8_MIN = np.iinfo(np.int8).min
_INT8_MAX = np.iinfo(np.int8).max
_INT16_MIN = np.iinfo(np.int16).min
_INT16_MAX = np.iinfo(np.int16).max
_INT32_MIN = np.iinfo(np.int32).min
_INT32_MAX = np.iinfo(np.int32).max

_FLOAT16_MIN = np.finfo(np.float16).min
_FLOAT16_MAX = np.finfo(np.float16).max
_FLOAT32_MIN = np.finfo(np.float32).min
_FLOAT32_MAX = np.finfo(np.float32).max


def compress_dataset(data: pd.DataFrame) -> None:
    """Compress the dataset.

    Args:
        data (pd.DataFrame): The dataset to compress.
    """
    for col in data.columns:
        col_dtype = data[col][:100].dtype

        if col_dtype != "object":
            col_series = data[col]
            col_min = col_series.min()
            col_max = col_series.max()

            if col_dtype == "float64":
                if (col_min > _FLOAT16_MIN) and (col_max < _FLOAT16_MAX):
                    data[col] = data[col].astype(np.float16)

                elif (col_min > _FLOAT32_MIN) and (col_max < _FLOAT32_MAX):
                    data[col] = data[col].astype(np.float32)

                else:
                    pass

            if col_dtype == "int64":
                if (col_min > _INT8_MIN / 2) and (col_max < _INT8_MAX / 2):
                    data[col] = data[col].astype(np.int8)
                elif (col_min > _INT16_MIN) and (col_max < _INT16_MAX):
                    data[col] = data[col].astype(np.int16)
                elif (col_min > _INT32_MIN) and (col_max < _INT32_MAX):
                    data[col] = data[col].astype(np.int32)
                else:
                    pass



# -*- coding: utf-8 -*-
"""CoRE Dataset."""
import os
from typing import Collection, Optional, Tuple

import numpy as np
import pandas as pd
from loguru import logger

from mofdscribe.constants import MOFDSCRIBE_PYSTOW_MODULE
from mofdscribe.datasets.checks import check_all_file_exists, length_check
from mofdscribe.datasets.dataset import AbstractStructureDataset
from mofdscribe.datasets.utils import compress_dataset

__all__ = ["CoREDataset"]


class CoREDataset(AbstractStructureDataset):
    """Dataset of gas uptake related features for a subset of CoRE MOFs.

    The labels were computed by Moosavi et al. (2020) [Moosavi2020]_.
    The raw labels and structures can accessed also on
    `MaterialsCloud <https://archive.materialscloud.org/record/2020.67>`_.

    To reduce the risk of data leakage, we  (by default) also only keep one representative
    structure for a "base refcode" (i.e. the first five letters of a refcode).
    For instance, the base refcode for IGAHED001 is IGAHED. Structures with same
    base refcode but different refcodes are often different refinements, or measurements
    at different temperatures and hence chemically quite similar. For instance,
    the base refcode `UMODEH` would appear 21 times, `KEDJAG` 17 times, and `UMOYOM` 17 times
    in the CoRE dataset used by Moosavi et al.
    Additionally, we (by default) only keep one structure per "structure hash"
    which is an approximate graph-isomoprhism check, assuming the VESTA bond thresholds
    for the derivation of the structure graph (e.g. the structure
    graph of ULOMAL occurs 59 in the CoRE database used by Moosavi et al.).

    .. warning::
        Even though we performed some basic sanity checks, there are currently still
        some structures that might chemically not be reasonable.
        Also, even though we only keep one structure per base refcode, there is still
        potential for data leakge. We urge users to still drop duplicates (or close neighbors)
        after featurization.

        If this set is used as test set, make sure to drop all overlapping entries in your training set.

    The years refer to the publication dates of the paper crossreferenced
    in the CSD entry of the structure. We excluded structures that are not
    deposited in the CSD.

    The available labels are:

        * 'pure_CO2_kH': Henry coefficient of CO2 obtained by Widom method in mol kg-1 Pa-1
        * 'pure_CO2_widomHOA': Heat of adsorption of CO2 obtained by Widom method in
        * 'pure_methane_kH': Henry coefficient of methane obtained by Widom method in mol kg-1 Pa-1
        * 'pure_methane_widomHOA': Heat of adsorption of methane obtained by Widom method
        * 'pure_uptake_CO2_298.00_15000': Pure CO2 uptake at 298.00 K and 15000 Pa in mol kg-1
        * 'pure_uptake_CO2_298.00_1600000': Pure CO2 uptake at 298.00 K and 1600000 Pa in mol kg-1
        * 'pure_uptake_methane_298.00_580000': Pure methane uptake at 298.00 K and 580000 Pa in mol kg-1
        * 'pure_uptake_methane_298.00_6500000': Pure methane uptake at 298.00 K and 6500000 Pa in mol kg-1
        * 'logKH_CO2': Logarithm of Henry coefficient of CO2 obtained by Widom method
        * 'logKH_CH4': Logarithm of Henry coefficient of methane obtained by Widom method
        * 'CH4DC': CH4 deliverable capacity in vSTP/v
        * 'CH4HPSTP': CH4 high pressure uptake in standard temperature and pressure in vSTP/v
        * 'CH4LPSTP': CH4 low pressure uptake in standard temperature and pressure in vSTP/v

    References:
        .. [Moosavi2020] Moosavi, S. M.; Nandy, A.; Jablonka, K. M.; Ongari, D.; Janet, J. P.; Boyd, P. G.; Lee,
            Y.; Smit, B.; Kulik, H. J. Understanding the Diversity of the Metal-Organic Framework Ecosystem.
            Nature Communications 2020, 11 (1), 4068. https://doi.org/10.1038/s41467-020-17755-8.

    """

    _files = {
        "v0.0.1": {
            "df": "https://zenodo.org/record/7032358/files/data.json?download=1",
            "structures": "https://zenodo.org/record/7032358/files/structures.tar.gz?download=1",
            "expected_length": 5393,
        }
    }

    def __init__(
        self,
        version: str = "v0.0.1",
        drop_basename_duplicates: bool = True,
        drop_graph_duplicates: bool = True,
        subset: Optional[Collection[int]] = None,
        drop_nan: bool = True,
    ):
        """Construct an instance of the CoRE dataset.

        Args:
            version (str): version number to use.
                Defaults to "v0.0.1".
            drop_basename_duplicates (bool): If True, keep only one structure
                per CSD basename. Defaults to True.
            drop_graph_duplicates (bool): If True, keep only one structure
                per decorated graph hash. Defaults to True.
            subset (Collection[int], optional): indices of the structures to include.
                Defaults to None.
            drop_nan (bool): If True, drop rows with NaN values in features or hashes.
                Defaults to True.

        Raises:
            ValueError: If the provided version number is not available.
        """
        self._drop_basename_duplicates = drop_basename_duplicates
        self._drop_nan = drop_nan
        self._drop_graph_duplicates = drop_graph_duplicates
        if version not in self._files:
            raise ValueError(
                f"Version {version} not available. Available versions: {list(self._files.keys())}"
            )
        self.version = version

        self._structure_dir = MOFDSCRIBE_PYSTOW_MODULE.ensure_untar(
            "CoRE",
            self.version,
            name="structures.tar.gz",
            url=self._files[version]["structures"],
        )

        self._df = pd.DataFrame(
            MOFDSCRIBE_PYSTOW_MODULE.ensure_json(
                "CoRE", self.version, name="data.json", url=self._files[version]["df"]
            )
        ).reset_index(drop=True)

        compress_dataset(self._df)

        length_check(self._df, self._files[version]["expected_length"])

        if drop_basename_duplicates:
            old_len = len(self._df)
            self._df = self._df.drop_duplicates(subset=["info.basename"])
            logger.debug(
                f"Dropped {old_len - len(self._df)} duplicate basenames. New length {len(self._df)}"
            )
        if drop_graph_duplicates:
            old_len = len(self._df)
            self._df = self._df.drop_duplicates(subset=["info.decorated_graph_hash"])
            logger.debug(
                f"Dropped {old_len - len(self._df)} duplicate graphs. New length {len(self._df)}"
            )
        self._df = self._df.reset_index(drop=True)
        if drop_nan:
            self._df.dropna(
                subset=[c for c in self._df.columns if c.startswith("features.")]
                + [c for c in self._df.columns if c.startswith("info.")],
                inplace=True,
            )
            self._df.reset_index(drop=True, inplace=True)

        if subset is not None:
            self._df = self._df.iloc[subset]
            self._df = self._df.reset_index(drop=True)

        self._structures = [
            os.path.join(self._structure_dir, f + ".cif") for f in self._df["info.name"]
        ]

        check_all_file_exists(self._structures)

        self._years = self._df["info.year"].values
        self._decorated_graph_hashes = self._df["info.decorated_graph_hash"].values
        self._undecorated_graph_hashes = self._df["info.undecorated_graph_hash"].values
        self._decorated_scaffold_hashes = self._df["info.decorated_scaffold_hash"].values
        self._undecorated_scaffold_hashes = self._df["info.undecorated_scaffold_hash"].values
        self._densities = self._df["info.density"].values
        self._labelnames = (c for c in self._df.columns if c.startswith("outputs."))
        self._featurenames = (c for c in self._df.columns if c.startswith("features."))
        self._infonames = (c for c in self._df.columns if c.startswith("info."))

    def get_subset(self, indices: Collection[int]) -> "AbstractStructureDataset":
        """Get a subset of the dataset.

        Args:
            indices (Collection[int]): indices of the structures to include.

        Returns:
            AbstractStructureDataset: a new dataset containing only the structures
                specified by the indices.
        """
        return CoREDataset(
            version=self.version,
            drop_basename_duplicates=self._drop_basename_duplicates,
            drop_graph_duplicates=self._drop_graph_duplicates,
            subset=indices,
            drop_nan=self._drop_nan,
        )

    @property
    def available_info(self) -> Tuple[str]:
        return self._infonames

    @property
    def available_features(self) -> Tuple[str]:
        return self._featurenames

    @property
    def available_labels(self) -> Tuple[str]:
        return self._labelnames

    def get_labels(self, idx: Collection[int], labelnames: Collection[str] = None) -> np.ndarray:
        labelnames = labelnames if labelnames is not None else self._labelnames
        return self._df.iloc[idx][list(labelnames)].values

    @property
    def citations(self) -> Tuple[str]:
        return [
            "@article{Moosavi2020,"
            "doi = {10.1038/s41467-020-17755-8},"
            "url = {https://doi.org/10.1038/s41467-020-17755-8},"
            "year = {2020},"
            "month = aug,"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {11},"
            "number = {1},"
            "author = {Seyed Mohamad Moosavi and Aditya Nandy and Kevin Maik Jablonka "
            "and Daniele Ongari and Jon Paul Janet and Peter G. Boyd and Yongjin Lee "
            "and Berend Smit and Heather J. Kulik},"
            "title = {Understanding the diversity of the metal-organic framework ecosystem},"
            "journal = {Nature Communications}"
            "}",
            "@article{Chung2019,"
            "doi = {10.1021/acs.jced.9b00835},"
            "url = {https://doi.org/10.1021/acs.jced.9b00835},"
            "year = {2019},"
            "month = nov,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {64},"
            "number = {12},"
            "pages = {5985--5998},"
            "author = {Yongchul G. Chung and Emmanuel Haldoupis and Benjamin J. Bucior "
            "and Maciej Haranczyk and Seulchan Lee and Hongda Zhang and "
            "Konstantinos D. Vogiatzis and Marija Milisavljevic and Sanliang Ling "
            "and Jeffrey S. Camp and Ben Slater and J. Ilja Siepmann and "
            "David S. Sholl and Randall Q. Snurr},"
            "title = {Advances,  Updates,  and Analytics for the Computation-Ready, "
            "Experimental Metal{\textendash}Organic Framework Database: {CoRE} {MOF} 2019},"
            r"journal = {Journal of Chemical {\&}amp$\mathsemicolon$ Engineering Data}"
            "}",
            "@article{Chung2014,"
            "doi = {10.1021/cm502594j},"
            "url = {https://doi.org/10.1021/cm502594j},"
            "year = {2014},"
            "month = oct,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {26},"
            "number = {21},"
            "pages = {6185--6192},"
            "author = {Yongchul G. Chung and Jeffrey Camp and "
            "Maciej Haranczyk and Benjamin J. Sikora and Wojciech Bury "
            "and Vaiva Krungleviciute and Taner Yildirim and Omar K. Farha "
            "and David S. Sholl and Randall Q. Snurr},"
            "title = {Computation-Ready,  Experimental Metal{\textendash}Organic Frameworks: "
            "A Tool To Enable High-Throughput Screening of Nanoporous Crystals},"
            "journal = {Chemistry of Materials}"
            "}",
        ]



# -*- coding: utf-8 -*-
"""Structures from the ARC-MOF dataset."""
import os
from typing import Collection, Optional, Tuple

import numpy as np
import pandas as pd
from loguru import logger

from mofdscribe.constants import MOFDSCRIBE_PYSTOW_MODULE
from mofdscribe.datasets.checks import check_all_file_exists, length_check
from mofdscribe.datasets.dataset import AbstractStructureDataset
from mofdscribe.datasets.utils import compress_dataset

__all__ = ["ARCMOFDataset"]


class ARCMOFDataset(AbstractStructureDataset):
    """
    Implements access to a subset of structures and labels of the ARC-MOF dataset [Burner2022]_.

    The subset consistes of the structures for which the authors reported
    process properties and for which we could compute graph hashes
    and features.

    .. warning::

        ARC-MOF is  "a database of ~280,000 MOFs which have been either
        experimentally characterized or computationally generated,
        spanning all publicly available MOF databases" [Burner2022]_.

        Therefore, there will be significant overlap with the other datasets.

    References:
        .. [Burner2022] `Burner, J.; Luo, J.; White, A.;
            Mirmiran, A.; Kwon, O.; Boyd, P. G.;
            Maley, S.; Gibaldi, M.; Simrod, S.;
            Ogden, V.; Woo, T. K.
            ChemRxiv 2022 <https://chemrxiv.org/engage/chemrxiv/article-details/62e04636cf661270d7b615c1>`_
    """

    _files = {
        "v0.0.1": {
            "df": "https://zenodo.org/record/7032350/files/data.json?download=1",
            "structures": "https://zenodo.org/record/7032350/files/structures.tar.gz?download=1",
            "expected_length": 22452,
        }
    }

    def __init__(
        self,
        version: str = "v0.0.1",
        drop_graph_duplicates: bool = True,
        subset: Optional[Collection[int]] = None,
    ):
        """Construct an instance of the CoRE dataset.

        Args:
            version (str): version number to use.
                Defaults to "v0.0.1".
            drop_graph_duplicates (bool): If True, keep only one structure
                per decorated graph hash. Defaults to True.
            subset (Collection[int], optional): indices of the structures to include.
                Defaults to None.

        Raises:
            ValueError: If the provided version number is not available.
        """
        self._drop_graph_duplicates = drop_graph_duplicates
        if version not in self._files:
            raise ValueError(
                f"Version {version} not available. Available versions: {list(self._files.keys())}"
            )
        self.version = version

        self._structure_dir = MOFDSCRIBE_PYSTOW_MODULE.ensure_untar(
            "ARCMOF",
            self.version,
            name="structures.tar.gz",
            url=self._files[version]["structures"],
        )

        self._df = pd.DataFrame(
            MOFDSCRIBE_PYSTOW_MODULE.ensure_json(
                "ARCMOF", self.version, name="data.json", url=self._files[version]["df"]
            )
        ).reset_index(drop=True)
        compress_dataset(self._df)

        length_check(self._df, self._files[version]["expected_length"])

        if drop_graph_duplicates:
            old_len = len(self._df)
            self._df = self._df.drop_duplicates(subset=["info.decorated_graph_hash"])
            logger.debug(
                f"Dropped {old_len - len(self._df)} duplicate graphs. New length {len(self._df)}"
            )
        self._df = self._df.reset_index(drop=True)
        if subset is not None:
            self._df = self._df.iloc[subset]
            self._df = self._df.reset_index(drop=True)

        self._structures = [
            os.path.join(self._structure_dir, f + ".cif") for f in self._df["info.name"]
        ]

        check_all_file_exists(self._structures)

        self._decorated_graph_hashes = self._df["info.decorated_graph_hash"].values
        self._undecorated_graph_hashes = self._df["info.undecorated_graph_hash"].values
        self._decorated_scaffold_hashes = self._df["info.decorated_scaffold_hash"].values
        self._undecorated_scaffold_hashes = self._df["info.undecorated_scaffold_hash"].values
        self._densities = self._df["info.density"].values
        self._labelnames = (c for c in self._df.columns if c.startswith("outputs."))
        self._featurenames = (c for c in self._df.columns if c.startswith("features."))
        self._infonames = (c for c in self._df.columns if c.startswith("info."))

    def get_subset(self, indices: Collection[int]) -> "AbstractStructureDataset":
        """Get a subset of the dataset.

        Args:
            indices (Collection[int]): indices of the structures to include.

        Returns:
            AbstractStructureDataset: a new dataset containing only the structures
                specified by the indices.
        """
        return ARCMOFDataset(
            version=self.version,
            drop_graph_duplicates=self._drop_graph_duplicates,
            subset=indices,
        )

    @property
    def available_info(self) -> Tuple[str]:
        return self._infonames

    @property
    def available_features(self) -> Tuple[str]:
        return self._featurenames

    @property
    def available_labels(self) -> Tuple[str]:
        return self._labelnames

    def get_labels(self, idx: Collection[int], labelnames: Collection[str] = None) -> np.ndarray:
        labelnames = labelnames if labelnames is not None else self._labelnames
        return self._df.iloc[idx][list(labelnames)].values

    @property
    def citations(self) -> Tuple[str]:
        return [
            "@article{Burner_2022,"
            "doi = {10.26434/chemrxiv-2022-mvr06},"
            "url = {https://doi.org/10.26434%2Fchemrxiv-2022-mvr06},"
            "year = 2022,"
            "month = {aug},"
            "publisher = {American Chemical Society ({ACS})},"
            "author = {Jake Burner and Jun Luo and Andrew White "
            "and Adam Mirmiran and Ohmin Kwon and Peter G. Boyd "
            "and Stephen Maley and Marco Gibaldi and Scott Simrod "
            "and Victoria Ogden and Tom K. Woo},"
            "title = {{ARC}-{MOF}: A Diverse Database of "
            "Metal-Organic Frameworks with {DFT}-Derived Partial "
            "Atomic Charges and Descriptors for Machine Learning}}"
        ]



# -*- coding: utf-8 -*-
"""Decorators and mixins for extending the functionality of featurizers."""
from functools import partial
from typing import Type

from pymatgen.core import IMolecule, IStructure, Molecule, Structure


def add_operates_on(cls, type: Type):
    """Add the `operates_on` method to a featurizer class.

    This is useful for SBUFeaturizers that need to know what type of
    input they need to pass to the featurizer.

    Args:
        cls (object): The class to add the `operates_on` classmethod to.
        type (Type): The types that the featurizer can operate on.

    Returns:
        cls: The class with the `operates_on` classmethod added.
    """
    try:
        cls._accepted_types.append(type)
    except AttributeError:
        cls._accepted_types = [type]

    def operates_on(self):
        return self._accepted_types

    cls.operates_on = operates_on
    return cls


operates_on_structure = partial(add_operates_on, type=Structure)
operates_on_istructure = partial(add_operates_on, type=IStructure)
operates_on_molecule = partial(add_operates_on, type=Molecule)
operates_on_imolecule = partial(add_operates_on, type=IMolecule)



from rdkit import Chem

from mofdscribe.featurizers.bu.smarts_matches import SmartsMatchCounter, number_smart_matches


def test_number_smart_matches():
    """Test that we get the correct number of matches."""
    mol = Chem.MolFromSmiles("C1=CC=CC=C1")
    smarts = ["n1nnnc1"]
    assert number_smart_matches(mol, smarts) == 0

    mol = Chem.MolFromSmiles("CCCP=O")
    smarts = ["[C,S,P]=O"]
    assert number_smart_matches(mol, smarts) == 1


def test_smart_match_counter(linker_molecule):
    """Test that we get the correct number of matches."""
    smarts = ["[S,P]=O"]
    featurizer = SmartsMatchCounter(smarts, feature_labels=["acid_groups"])
    assert featurizer.featurize(linker_molecule) == [0]

    aromatic_ring_smarts = ["c1ccccc1"]
    featurizer = SmartsMatchCounter(aromatic_ring_smarts, feature_labels=["aromatic_rings"])
    assert featurizer.featurize(linker_molecule) == [6]



# -*- coding: utf-8 -*-
"""Featurizers using persistent homology -- applied in an atom-centred manner."""
from collections import defaultdict
from typing import List, Optional, Tuple, Union

import numpy as np
from element_coder import encode_many
from matminer.featurizers.base import BaseFeaturizer
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer
from mofdscribe.featurizers.utils import flatten
from mofdscribe.featurizers.utils.aggregators import ARRAY_AGGREGATORS
from mofdscribe.featurizers.utils.extend import operates_on_istructure, operates_on_structure

from ._tda_helpers import construct_pds_cached, diagrams_to_bd_arrays, persistent_diagram_stats


# Todo: allow doing this with cutoff and coordination shells
# ToDo: check if this works with molecules
@operates_on_istructure
@operates_on_structure
class AtomCenteredPHSite(BaseFeaturizer):
    """Site featurizer for atom-centered statistics of persistence diagrams.

    This featurizer is an abstraction of the on described in the work of Jiang
    et al. (2021) [Jiang2021]_. It computes the persistence diagrams for the
    neighborhood of every site and then aggregates the diagrams by computing
    certain statistics.

    To use this featurizer on a complete structure without additional
    resolutions over the chemistry, you can wrap it in a
    :class:`~matminer.featurizers.structure.SiteStatsFingerprint`.

    Examples:
        >>> from matminer.featurizers.structure import SiteStatsFingerprint
        >>> from mofdscribe.topology.atom_centered_ph import AtomCenteredPHSite
        >>> featurizer = SiteStatsFingerprint(AtomCenteredPHSite())
        >>> featurizer.featurize(structure)
        np.array([2,...]) # np.array with features

    However, if you want the additional chemical dimension, you can use the the
    :class:`~mofdscribe.topology.atom_centered_ph.AtomCenteredPH`.
    """

    def __init__(
        self,
        aggregation_functions: Tuple[str] = ("min", "max", "mean", "std"),
        cutoff: float = 12,
        dimensions: Tuple[int] = (1, 2),
        alpha_weight: Optional[str] = None,
    ) -> None:
        """
        Construct a new AtomCenteredPHSite featurizer.

        Args:
            aggregation_functions (Tuple[str]): Aggregations to
                compute on the persistence diagrams (over birth/death time and
                persistence). Defaults to ("min", "max", "mean", "std").
            cutoff (float): Consider neighbors of site within this radius (in
                Angstrom). Defaults to 12.
            dimensions (Tuple[int]): Betti numbers of consider.
                0 describes isolated components, 1 cycles and 2 cavities.
                Defaults to (1, 2).
            alpha_weight (Optional[str]):  If specified, the use weighted alpha shapes,
                i.e., replacing the points with balls of varying radii.
                For instance `atomic_radius_calculated` or `van_der_waals_radius`.
        """
        self.aggregation_functions = aggregation_functions
        self.cutoff = cutoff
        self.dimensions = dimensions
        self.alpha_weight = alpha_weight

    def featurize(self, s: Union[Structure, IStructure], idx: int) -> np.ndarray:
        neighbors = s.get_neighbors(s[idx], self.cutoff)
        neighbor_structure = IStructure.from_sites(neighbors)
        if self.alpha_weight is not None:
            weights = encode_many(
                [str(s.symbol) for s in neighbor_structure.species], self.alpha_weight
            )
        else:
            weights = None

        diagrams = construct_pds_cached(neighbor_structure.cart_coords, weights=weights)

        diagrams = diagrams_to_bd_arrays(diagrams)

        results = {}
        for dim in self.dimensions:
            key = f"dim{dim}"
            results[key] = persistent_diagram_stats(diagrams[key], self.aggregation_functions)
        return np.array(list(flatten(results).values()))

    def _get_feature_labels(self) -> List[str]:
        names = []
        for dim in self.dimensions:
            dim_key = f"dim{dim}"
            for parameter in ("birth", "death", "persistence"):
                for aggregation in self.aggregation_functions:
                    names.append(f"{dim_key}_{parameter}_{aggregation}")
        return names

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka"]

    def citations(self) -> List[str]:
        return [
            "@article{Jiang2021,"
            "doi = {10.1038/s41524-021-00493-w},"
            "url = {https://doi.org/10.1038/s41524-021-00493-w},"
            "year = {2021},"
            "month = feb,"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {7},"
            "number = {1},"
            "author = {Yi Jiang and Dong Chen and Xin Chen and Tangyi Li "
            "and Guo-Wei Wei and Feng Pan},"
            "title = {Topological representations of crystalline compounds "
            "for the machine-learning prediction of materials properties},"
            "journal = {npj Computational Materials}"
            "}",
            "@article{doi:10.1021/acs.jpcc.0c01167,"
            "author = {Krishnapriyan, Aditi S. and Haranczyk, Maciej and Morozov, Dmitriy},"
            "title = {Topological Descriptors Help Predict Guest Adsorption in Nanoporous Materials},"
            "journal = {The Journal of Physical Chemistry C},"
            "volume = {124},"
            "number = {17},"
            "pages = {9360-9368},"
            "year = {2020},"
            "doi = {10.1021/acs.jpcc.0c01167},"
            "}",
        ]


# ToDo: Leverage symmetry to do not recompute for symmetry-equivalent sites
class AtomCenteredPH(MOFBaseFeaturizer):
    """Atom-centered featurizer for persistence diagrams.

    It runs :class:`~mofdscribe.topology.atom_centered_ph.AtomCenteredPH` for every site.

    It aggregates the results over atom types that are specified in the constructor
    via aggregation functions specified in the constructor.
    """

    def __init__(
        self,
        atom_types: Tuple[str] = (
            "C-H-N-O",
            "F-Cl-Br-I",
            "Cu-Mn-Ni-Mo-Fe-Pt-Zn-Ca-Er-Au-Cd-Co-Gd-Na-Sm-Eu-Tb-V"
            "-Ag-Nd-U-Ba-Ce-K-Ga-Cr-Al-Li-Sc-Ru-In-Mg-Zr-Dy-W-Yb-Y-"
            "Ho-Re-Be-Rb-La-Sn-Cs-Pb-Pr-Bi-Tm-Sr-Ti-Hf-Ir-Nb-Pd-Hg-"
            "Th-Np-Lu-Rh-Pu",
        ),
        compute_for_all_elements: Optional[bool] = True,
        aggregation_functions: Tuple[str] = ("min", "max", "mean", "std"),
        species_aggregation_functions: Tuple[str] = ("min", "max", "mean", "std"),
        cutoff: float = 12,
        dimensions: Tuple[int] = (1, 2),
        primitive: bool = False,
        alpha_weight: Optional[str] = None,
    ) -> None:
        """
        Construct a new AtomCenteredPH featurizer.

        Args:
            atom_types (tuple): Atoms that are used to create substructures
                that are analysed using persistent homology.
                If multiple atom types separated by hash are provided, e.g. "C-H-N-O",
                then the substructure consists of all atoms of type C, H, N, or O.
                Defaults to ( "C-H-N-O", "F-Cl-Br-I",
                "Cu-Mn-Ni-Mo-Fe-Pt-Zn-Ca-Er-Au-Cd-Co-Gd-Na-Sm-Eu-Tb-V-Ag-Nd-U-Ba-Ce-K-Ga-
                "Cr-Al-Li-Sc-Ru-In-Mg-Zr-Dy-W-Yb-Y-Ho-Re-Be-Rb-La-Sn-Cs-Pb-Pr-Bi-Tm-Sr-Ti-Hf-Ir-
                "Nb-Pd-Hg-Th-Np-Lu-Rh-Pu", ).
            compute_for_all_elements (bool): Compute descriptor for original structure with all atoms.
                Defaults to True.
            aggregation_functions (Tuple[str]): Aggregations to compute on the persistence
                diagrams (over birth/death time and persistence).
                Defaults to ("min", "max", "mean", "std").
            species_aggregation_functions (Tuple[str]): Aggregations to use to combine
                features derived for sites of a specific atom type, e.g., the site features of all `C-H-N-O`.
                Defaults to ("min", "max", "mean", "std").
            cutoff (float): Consider neighbors of site within this radius (in Angstrom).
                Defaults to 12.
            dimensions (Tuple[int]): Betti numbers of consider. 0 describes isolated components,
                1 cycles and 2 cavities. Defaults to (1, 2).
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to False.
            alpha_weight (Optional[str]):  If specified, the use weighted alpha shapes,
                i.e., replacing the points with balls of varying radii.
                For instance `atomic_radius_calculated` or `van_der_waals_radius`.
        """
        self.aggregation_functions = aggregation_functions
        self.species_aggregation_functions = species_aggregation_functions
        self.cutoff = cutoff
        self.dimensions = dimensions
        self.alpha_weight = alpha_weight
        self.site_featurizer = AtomCenteredPHSite(
            aggregation_functions=aggregation_functions,
            cutoff=cutoff,
            dimensions=dimensions,
            alpha_weight=alpha_weight,
        )
        atom_types = [] if atom_types is None else atom_types
        self.atom_types = (
            list(atom_types) + ["all"] if compute_for_all_elements else list(atom_types)
        )
        self.compute_for_all_elements = compute_for_all_elements

        super().__init__(primitive=primitive)

    def _get_relevant_atom_type(self, element: str) -> str:
        for atom_type in self.atom_types:
            if element in atom_type:
                return atom_type

    def _featurize(self, s: Union[Structure, IStructure]) -> np.ndarray:
        results = defaultdict(list)
        for idx, site in enumerate(s):
            atom_type = self._get_relevant_atom_type(site.specie.symbol)
            features = self.site_featurizer.featurize(s, idx)
            if atom_type is not None:
                results[atom_type].append(features)
            results["all"].append(features)
        long_results = []
        for atom_type in self.atom_types:
            if atom_type not in results:
                long_results.extend(
                    np.zeros(
                        len(self.site_featurizer.feature_labels())
                        * len(self.species_aggregation_functions)
                    )
                )
            else:
                v = np.array(results[atom_type])
                for aggregation in self.species_aggregation_functions:
                    agg_func = ARRAY_AGGREGATORS[aggregation]
                    long_results.extend(agg_func(v, axis=0))

        return np.array(long_results)

    def _get_feature_labels(self) -> List[str]:
        names = []

        for atom_type in self.atom_types:
            for aggregation in self.species_aggregation_functions:
                for fl in self.site_featurizer.feature_labels():
                    names.append(f"{atom_type}_{aggregation}_{fl}")

        return names

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka", "Aditi Krishnapriyan"]

    def citations(self) -> List[str]:
        return [
            "@article{Jiang2021,"
            "doi = {10.1038/s41524-021-00493-w},"
            "url = {https://doi.org/10.1038/s41524-021-00493-w},"
            "year = {2021},"
            "month = feb,"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {7},"
            "number = {1},"
            "author = {Yi Jiang and Dong Chen and Xin Chen and Tangyi Li and Guo-Wei Wei and Feng Pan},"
            "title = {Topological representations of crystalline compounds for "
            "the machine-learning prediction of materials properties},"
            "journal = {npj Computational Materials}"
            "}",
            "@article{doi:10.1021/acs.jpcc.0c01167,"
            "author = {Krishnapriyan, Aditi S. and Haranczyk, Maciej and Morozov, Dmitriy},"
            "title = {Topological Descriptors Help Predict Guest Adsorption in Nanoporous Materials},"
            "journal = {The Journal of Physical Chemistry C},"
            "volume = {124},"
            "number = {17},"
            "pages = {9360-9368},"
            "year = {2020},"
            "doi = {10.1021/acs.jpcc.0c01167},"
            "}",
        ]



# -*- coding: utf-8 -*-
"""Test BW dataset."""
import numpy as np
from pymatgen.core import IStructure

from mofdscribe.datasets.bw_dataset import BWDataset


def test_bw():
    """Ensure we can instantiate the BW dataset and access a few key methods."""
    bw = BWDataset()
    assert isinstance(list(bw.get_structures([1]))[0], IStructure)
    assert isinstance(bw.get_labels([1]), np.ndarray)

    # make sure we can get a subset of the dataset
    bw_subset = bw.get_subset([1, 2, 3, 8])
    assert isinstance(bw_subset, BWDataset)
    assert len(bw_subset._df) == 4
    assert (
        bw._df.iloc[[1, 2, 3, 8]]["info.name"].values == bw_subset._df["info.name"].values
    ).all()
    assert list(bw.get_structures([1]))[0] == list(bw_subset.get_structures([0]))[0]



# -*- coding: utf-8 -*-
"""Test ARABG dataset."""
import numpy as np
from pymatgen.core import IStructure

from mofdscribe.datasets.arabg_dataset import ARABGDataset


def test_arabg():
    """Ensure we can instantiate the ARABG dataset and access a few key methods."""
    arabg = ARABGDataset()
    assert isinstance(list(arabg.get_structures([1]))[0], IStructure)
    assert isinstance(arabg.get_labels([1]), np.ndarray)

    # make sure we can get a subset of the dataset
    arabg_subset = arabg.get_subset([1, 2, 3, 8])
    assert isinstance(arabg_subset, ARABGDataset)
    assert len(arabg_subset._df) == 4
    assert (
        arabg._df.iloc[[1, 2, 3, 8]]["info.name"].values == arabg_subset._df["info.name"].values
    ).all()
    assert list(arabg.get_structures([1]))[0] == list(arabg_subset.get_structures([0]))[0]



# -*- coding: utf-8 -*-
"""Revised autocorrelation functions (RACs) for MOFs."""
from collections import OrderedDict, defaultdict
from typing import Collection, List, Optional, Tuple, Union

import numpy as np
from element_coder import encode
from loguru import logger
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer
from mofdscribe.featurizers.utils.aggregators import AGGREGATORS, ARRAY_AGGREGATORS
from mofdscribe.featurizers.utils.extend import operates_on_istructure, operates_on_structure
from mofdscribe.featurizers.utils.structure_graph import (
    get_connected_site_indices,
    get_neighbors_at_distance,
    get_structure_graph,
)

from ._fragment import get_bb_indices

__all__ = ("RACS",)


def _compute_racs(
    start_indices: Collection[int],
    structure_graph: StructureGraph,
    properties: Tuple[str],
    scope: int,
    property_aggregations: Tuple[str],
    corr_aggregations: Tuple[str],
    part_name: str = "",
    nan_value: float = np.nan,
):
    racs = defaultdict(lambda: defaultdict(list))
    if len(start_indices) == 0:
        logger.debug(f"No start indices for {part_name}")
        for prop in properties:
            for agg in property_aggregations:
                racs[prop][agg].append(nan_value)
    else:
        for start_atom in start_indices:
            _, neighbors = get_neighbors_at_distance(structure_graph, start_atom, scope)
            num_neighbors = len(neighbors)
            # We only branch if there are actually neighbors scope many bonds away ...

            if num_neighbors > 0:
                site = structure_graph.structure[start_atom]

                for neighbor in neighbors:

                    n = structure_graph.structure[neighbor]
                    for prop in properties:
                        if prop in ("I", 1):
                            p0 = 1
                            p1 = 1
                        elif prop == "T":
                            p0 = num_neighbors
                            p1 = len(get_connected_site_indices(structure_graph, neighbor))
                        else:
                            p0 = encode(site.specie.symbol, prop)
                            p1 = encode(n.specie.symbol, prop)

                        for agg in property_aggregations:
                            agg_func = AGGREGATORS[agg]
                            racs[prop][agg].append(agg_func((p0, p1)))
            else:
                logger.debug(f"No neighbors found for {start_atom}")
                for prop in properties:
                    for agg in property_aggregations:
                        racs[prop][agg].append(nan_value)
    aggregated_racs = {}

    for property_name, property_values in racs.items():
        for aggregation_name, aggregation_values in property_values.items():
            for corr_agg in corr_aggregations:
                agg_func = ARRAY_AGGREGATORS[corr_agg]
                name = f"racs_bb-{part_name}_prop-{property_name}_scope-{scope}_propagg-{aggregation_name}_corragg-{corr_agg}"  # noqa: E501
                aggregated_racs[name] = agg_func(aggregation_values)

    return aggregated_racs


def _get_racs_for_bbs(
    bb_indices: Collection[int],
    structure_graph: StructureGraph,
    properties: Tuple[str],
    scopes: List[int],
    property_aggregations: Tuple[str],
    corr_aggregations: Tuple[str],
    bb_aggregations: Tuple[str],
    bb_name: str = "",
):
    bb_racs = defaultdict(list)

    if not bb_indices:
        # one nested list to make it trigger filling it with nan values
        bb_indices = [[]]
    for start_indices in bb_indices:
        for scope in scopes:
            racs = _compute_racs(
                start_indices,
                structure_graph,
                properties,
                scope,
                property_aggregations,
                corr_aggregations,
                bb_name,
            )
            for k, v in racs.items():
                bb_racs[k].append(v)

    aggregated_racs = {}
    for racs_name, racs_values in bb_racs.items():
        for bb_agg in bb_aggregations:
            agg_func = ARRAY_AGGREGATORS[bb_agg]
            name = f"{racs_name}_bbagg-{bb_agg}"
            aggregated_racs[name] = agg_func(racs_values)

    return aggregated_racs


@operates_on_istructure
@operates_on_structure
class RACS(MOFBaseFeaturizer):
    r"""Modified version of the revised autocorrelation functions (RACs) for MOFs.

    Originally proposed by Moosavi et al. (10.1038/s41467-020-17755-8).
    In the original paper, RACs were computed as

    .. math::
        {}_{{\rm{scope}}}^{{\rm{start}}}{P}_{d}^{{\rm{diff}}}=\mathop{\sum }\limits_{i}^{{\rm{start}}}\mathop{\sum }\limits_{j}^{{\rm{scope}}}({P}_{i}-{P}_{j})\delta ({d}_{i,j},d). # noqa: E501

    Here, we allow to replace the double sum by different aggregations. We call
    this `corr_agg`. The default `sum` is equivalent to the original RACS.
    Moreover, the implementation here keeps track of different linker/node
    molecules and allows to compute and aggregate the RACS for each molecule
    separately. The `bb_agg` feature then determines how those RACs for each BB
    are aggregated. The `sum` is equivalent to the original RACS (i.e. all
    applicable linker atoms would be added to the start/scope lists).

    Note that the "bbs" define the start atoms. However, we still consider the whole
    neighborhood of each start atom (i.e., including atoms that are not part of this "bb").

    Furthermore, here we allow to allow any of the
    `element-coder <https://github.com/kjappelbaum/element-coder>`_ properties to be used as
    `property` :math:`P_{i}`.

    To use to original implementation, see `molSimplify
    <https://github.com/hjkgrp/molSimplify>`_.
    """

    def __init__(
        self,
        attributes: Tuple[Union[int, str]] = ("X", "mod_pettifor", "I", "T"),
        scopes: Tuple[int] = (1, 2, 3),
        prop_agg: Tuple[str] = ("product", "diff"),
        corr_agg: Tuple[str] = ("sum", "avg"),
        bb_agg: Tuple[str] = ("avg", "sum"),
        bond_heuristic: str = "vesta",
        bbs: Optional[Tuple[str]] = (
            "linker_all",
            "linker_connecting",
            "linker_functional",
            "linker_scaffold",
            "nodes",
        ),
        primitive: bool = True,
    ) -> None:
        """
        Initialize the RACS featurizer.

        Args:
            attributes (Tuple[Union[int, str]]): Properties that are correlated.
                Defaults to ("X", "mod_pettifor", "I", "T").
            scopes (Tuple[int]): Number of edges to traverse. Defaults to (1, 2, 3).
            prop_agg (Tuple[str]): Function for aggregating the properties.
                Defaults to ("product", "diff").
            corr_agg (Tuple[str]): Function to aggregate the properties across different start/scopes.
                Defaults to ("sum", "avg").
            bb_agg (Tuple[str]): Function used to aggregate the properties across different building blocks.
                Defaults to ("avg", "sum").
            bond_heuristic (str): Method used to guess bonds. Defaults to "vesta".
            bbs (Tuple[str]): Building blocks to use. Defaults to ("linker_all",
                "linker_connecting", "linker_functional", "linker_scaffold", "nodes").
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.
        """
        self.attributes = attributes
        self.scopes = scopes
        self.prop_agg = prop_agg
        self.corr_agg = corr_agg
        self.bb_agg = bb_agg
        self.bond_heuristic = bond_heuristic
        if bbs is not None:
            self._bbs = bbs
        else:
            self._bbs = [
                "linker_all",
                "linker_connecting",
                "linker_functional",
                "linker_scaffold",
                "nodes",
            ]
        super().__init__(primitive=primitive)

    def _featurize(self, structure: Union[Structure, IStructure]) -> np.ndarray:
        if isinstance(structure, Structure):
            structure = IStructure.from_sites(structure)
        sg = get_structure_graph(structure, self.bond_heuristic)

        racs = {}
        # This finds all indices for a particular subset of atoms
        # e.g. "nodes", "linker_all", "linker_connecting", "linker_functional", "linker_scaffold"
        bb_indices = get_bb_indices(sg)
        for bb in self._bbs:
            racs.update(
                _get_racs_for_bbs(
                    bb_indices[bb],
                    sg,
                    self.attributes,
                    self.scopes,
                    self.prop_agg,
                    self.corr_agg,
                    self.bb_agg,
                    bb,
                )
            )
        racs_ordered = OrderedDict(sorted(racs.items()))
        return np.array(list(racs_ordered.values()))

    def _get_feature_labels(self) -> List[str]:
        names = []
        for bb in self._bbs:
            for scope in self.scopes:
                for prop in self.attributes:
                    for property_agg in self.prop_agg:
                        for cor_agg in self.corr_agg:
                            for bb_agg in self.bb_agg:
                                names.append(
                                    f"racs_bb-{bb}_prop-{prop}_scope-{scope}_propagg-{property_agg}_corragg-{cor_agg}_bbagg-{bb_agg}"  # noqa: E501
                                )

        names = sorted(names)
        return names

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def citations(self) -> List[str]:
        return [
            "@article{Moosavi2020,"
            "doi = {10.1038/s41467-020-17755-8},"
            "url = {https://doi.org/10.1038/s41467-020-17755-8},"
            "year = {2020},"
            "month = aug,"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {11},"
            "number = {1},"
            "author = {Seyed Mohamad Moosavi and Aditya Nandy and "
            "Kevin Maik Jablonka and Daniele Ongari and Jon Paul Janet "
            "and Peter G. Boyd and Yongjin Lee and Berend Smit and Heather J. Kulik},"
            "title = {Understanding the diversity of the metal-organic framework ecosystem},"
            "journal = {Nature Communications}"
            "}",
            "@article{Janet2017,"
            "doi = {10.1021/acs.jpca.7b08750},"
            "url = {https://doi.org/10.1021/acs.jpca.7b08750},"
            "year = {2017},"
            "month = nov,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {121},"
            "number = {46},"
            "pages = {8939--8954},"
            "author = {Jon Paul Janet and Heather J. Kulik},"
            "title = {Resolving Transition Metal Chemical Space: "
            "Feature Selection for Machine Learning and Structure{\textendash}Property Relationships},"
            "journal = {The Journal of Physical Chemistry A}"
            "}",
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka"]



# -*- coding: utf-8 -*-
"""Test the atom centred PH featurizer."""
import numpy as np
import pytest

from mofdscribe.featurizers.topology.atom_centered_ph import AtomCenteredPH, AtomCenteredPHSite

from ..helpers import is_jsonable


def test_atom_centered_ph_site(hkust_structure, irmof_structure, cof_structure):
    """Ensure we get the correct number of features and that they are different for different sites."""
    for i, structure in enumerate([hkust_structure, irmof_structure, cof_structure]):
        featurizer = AtomCenteredPHSite()
        features = featurizer.featurize(structure, 0)
        feature_labels = featurizer.feature_labels()
        assert len(features) == len(feature_labels)
        features_1 = featurizer.featurize(structure, 1)
        assert len(features_1) == len(feature_labels)
        if i < 2:
            # The metals should be equivalent
            assert np.allclose(features, features_1)
            features_not_metal = featurizer.featurize(structure, -1)
            assert np.abs(features - features_not_metal).sum() > 0
    assert is_jsonable(dict(zip(featurizer.feature_labels(), features)))
    assert features.ndim == 1


def test_atom_centered_ph(hkust_structure, irmof_structure, hkust_la_structure):
    """Ensure we get the correct number of features and that they are different for different structures."""
    for structure in [hkust_structure]:
        featurizer = AtomCenteredPH()
        features = featurizer.featurize(structure)
        feature_labels = featurizer.feature_labels()
        assert len(features) == len(feature_labels)

    features_hkust = featurizer.featurize(hkust_structure)
    features_irmof = featurizer.featurize(irmof_structure)
    assert (features_hkust != features_irmof).any()
    assert is_jsonable(dict(zip(featurizer.feature_labels(), features)))
    assert features.ndim == 1

    # now, try that encoding chemistry with varying atomic radii works
    # we do this by computing the PH histograms for structures with same geometry
    # and connectivity, but different atoms
    featurizer = AtomCenteredPH(atom_types=None, alpha_weight="van_der_waals_radius")
    features_hkust = featurizer.featurize(hkust_structure)
    features_hkust_la = featurizer.featurize(hkust_la_structure)
    assert features_hkust.shape == features_hkust_la.shape
    assert (features_hkust != features_hkust_la).any()

    # # now, to be sure do not encode the same thing with the atomic radius
    featurizer = AtomCenteredPH(atom_types=None, alpha_weight=None)
    features_hkust = featurizer.featurize(hkust_structure)
    features_hkust_la = featurizer.featurize(hkust_la_structure)
    assert features_hkust.shape == features_hkust_la.shape
    assert features_hkust == pytest.approx(features_hkust_la, rel=1e-2)



# -*- coding: utf-8 -*-
"""In-dataset predictions for the logarithmitic CO2 Henry coefficients"""
from typing import Optional

import numpy as np

from mofdscribe.datasets import CoREDataset
from mofdscribe.splitters.splitters import DensitySplitter, HashSplitter

from .mofbench import MOFBenchRegression

__all__ = ["LogkHCO2IDBench", "LogkHCO2OODBench"]


class LogkHCO2IDBench(MOFBenchRegression):
    """Benchmarking models for the logarithmic CO2 Henry coefficient under in-domain conditions.

    In-distribution implies that we use a cluster stratified splitter
    that ensures that the ratios of different clusters in the training
    and test set are the same.
    """

    def __init__(
        self,
        model,
        name: str,
        version: Optional[str] = "v0.0.1",
        features: Optional[str] = None,
        model_type: Optional[str] = None,
        reference: Optional[str] = None,
        implementation: Optional[str] = None,
        debug: bool = False,
        patch_in_ds: bool = False,
    ):
        """Initialize the  log KH CO2 interpolation benchmark.

        Args:
            model (object): The model to be benchmarked.
                Must implement the `fit` and `predict` methods.
            name (str): The name of the modeling approach.
            version (str, optional): Version of the dataset to use.
                Defaults to "v0.0.1".
            features (str, optional): Description of the features used in the model.
                Defaults to None.
            model_type (str, optional): Model type (e.g. Conv-Net, BERT, XGBoost).
                Defaults to None.
            reference (str, optional): Reference with more details about modeling approach.
                Defaults to None.
            implementation (str, optional): Link to implementation. Defaults to None.
            debug (bool): If True, use a small dataset (1% of full dataset) for debugging.
                Defaults to False.
            patch_in_ds (bool): If True, the dataset will be patched into the model class
                under the `ds` attribute.
        """
        super().__init__(
            model,
            ds=CoREDataset(version),
            splitter=HashSplitter(
                CoREDataset(version),
                stratification_col="outputs.logKH_CO2",
                sample_frac=0.01 if debug else 1.0,
            ),
            target=["outputs.logKH_CO2"],
            task="logKH_CO2_id",
            k=5,
            version=version,
            features=features,
            name=name,
            model_type=model_type,
            reference=reference,
            implementation=implementation,
            debug=debug,
            patch_in_ds=patch_in_ds,
        )


class LogkHCO2OODBench(MOFBenchRegression):
    """Benchmarking models for the logarithmic CO2 Henry coefficient under "out-of-domain" conditions.

    "Out-of-domain" conditions means that every of the 5 training fold will only see 4 out of the 5
    quantile bins.
    This implies that 2 runs are extrapolative and the other 3 need to "fill holes in the distribution".
    """

    def __init__(
        self,
        model,
        version: Optional[str] = "v0.0.1",
        features: Optional[str] = None,
        name: Optional[str] = None,
        model_type: Optional[str] = None,
        reference: Optional[str] = None,
        implementation: Optional[str] = None,
        debug: bool = False,
        patch_in_ds: bool = False,
    ):
        """Initialize the  log KH CO2 extrapolation benchmark.

        Args:
            model (object): The model to be benchmarked.
                Must implement the `fit` and `predict` methods.
            name (str): The name of the modeling approach.
            version (str, optional): Version of the dataset to use.
                Defaults to "v0.0.1".
            features (str, optional): Description of the features used in the model.
                Defaults to None.
            model_type (str, optional): Model type (e.g. Conv-Net, BERT, XGBoost).
                Defaults to None.
            reference (str, optional): Reference with more details about modeling approach.
                Defaults to None.
            implementation (str, optional): Link to implementation. Defaults to None.
            debug (bool): If True, use a small dataset (1% of full dataset) for debugging.
                Defaults to False.
            patch_in_ds (bool): If True, the dataset will be patched into the model class
                under the `ds` attribute.
        """
        super().__init__(
            model,
            ds=CoREDataset(version),
            splitter=DensitySplitter(
                CoREDataset(version),
                sample_frac=0.01 if debug else 1.0,
                density_q=np.linspace(0, 1, 6),
            ),
            target=["outputs.logKH_CO2"],
            task="logKH_CO2_ood",
            k=5,
            version=version,
            features=features,
            name=name,
            model_type=model_type,
            reference=reference,
            implementation=implementation,
            debug=debug,
            patch_in_ds=patch_in_ds,
        )



# -*- coding: utf-8 -*-
"""Describe the chemical composition of structures."""
from collections import defaultdict
from typing import List, Tuple, Union

import numpy as np
from element_coder import encode
from matminer.featurizers.base import BaseFeaturizer
from pymatgen.core import IMolecule, IStructure, Molecule, Structure

from mofdscribe.featurizers.utils.aggregators import ARRAY_AGGREGATORS
from mofdscribe.featurizers.utils.extend import (
    operates_on_imolecule,
    operates_on_istructure,
    operates_on_molecule,
    operates_on_structure,
)


@operates_on_molecule
@operates_on_imolecule
@operates_on_istructure
@operates_on_structure
class CompositionStats(BaseFeaturizer):
    """
    Describe the composition of molecules by computing statistics of their compositions.

    The featurizer will encode the element on all sites in the structure
    using user-defined encodings. Then it aggregates those encodings using
    user-defined encodings (e.g. min, max, min).
    """

    def __init__(
        self,
        encodings: Tuple[str] = ("mod_pettifor", "X"),
        aggregations: Tuple[str] = ("mean", "std", "max", "min"),
    ) -> None:
        """Initialize a CompositionStats featurizer.

        Args:
            encodings (Tuple[str]): Encoding used for the elements.
                Can be one of :py:obj:`element_coder.data.coding_data._PROPERTY_KEYS`.
                Defaults to ("mod_pettifor", "X").
            aggregations (Tuple[str]): Statistic to compute over the element encodings.
                Can be one of :py:obj:`mofdscribe.featurizers.utils.aggregators.ARRAY_AGGREGATORS`.
                Defaults to ("mean", "std", "max", "min").
        """
        self.aggregations = aggregations
        self.encodings = encodings

    def feature_labels(self) -> List[str]:
        feature_labels = []

        for encoding in self.encodings:
            for agg in self.aggregations:
                feature_labels.append(f"composition_stats_{encoding}_{agg}")

        return feature_labels

    def featurize(self, molecule: Union[Molecule, IMolecule, Structure, IStructure]) -> np.ndarray:
        encodings = defaultdict(list)
        for encoding in self.encodings:
            for site in molecule.sites:
                encodings[encoding] = encode(site.specie, encoding)

        features = []
        for encoding in self.encodings:
            for agg in self.aggregations:
                features.append(ARRAY_AGGREGATORS[agg](encodings[encoding]))

        return np.array(features)

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka"]

    def citations(self) -> List[str]:
        return []



# -*- coding: utf-8 -*-
"""Test thermal stability dataset."""
import numpy as np
import pytest
from pymatgen.core import IStructure

from mofdscribe.datasets.thermal_stability_dataset import ThermalStabilityDataset


@pytest.mark.xdist_group(name="thermal-stability-ds")
def test_thermal_stability():
    dataset = ThermalStabilityDataset()

    assert isinstance(list(dataset.get_structures([1]))[0], IStructure)
    assert isinstance(dataset.get_labels([1]), np.ndarray)
    assert isinstance(dataset.get_years([1]), np.ndarray)
    assert len(dataset.get_years([1])) == 1
    assert len(dataset.get_years([1, 2, 4])) == 3

    # make sure we can get a subset of the dataset
    subset = dataset.get_subset([1, 2, 3, 8])
    assert isinstance(subset, ThermalStabilityDataset)
    assert len(subset._df) == 4
    assert (
        dataset._df.iloc[[1, 2, 3, 8]]["info.basename"].values == subset._df["info.basename"].values
    ).all()
    assert list(dataset.get_structures([1]))[0] == list(subset.get_structures([0]))[0]



# -*- coding: utf-8 -*-
"""Feature MOFs using economic descriptors."""
import operator
import os
from collections import OrderedDict
from typing import List, Tuple

import numpy as np
import pandas as pd

from mofdscribe.featurizers.base import MOFBaseFeaturizer

_THIS_DIR = os.path.dirname(os.path.abspath(__file__))

_ELEMENT_PRICES = pd.read_csv(os.path.join(_THIS_DIR, "data", "elementprices.csv"), comment="#")


def _get_element_prices():
    """Return element prices dictionary."""
    prices_dict = {}

    for _i, row in _ELEMENT_PRICES.iterrows():
        # clean 'price' column
        price = row["Price [USD/kg]"]
        if isinstance(price, str) and "–" in price:
            prices = price.split("–")
            price = (float(prices[0]) + float(prices[1])) / 2
        elif price in ["Not traded.", "No reliable price available."]:
            price = np.nan
        elif price == "Negative price.":
            price = 0

        prices_dict[row["Symbol"]] = float(price)

    return prices_dict


def _get_element_abundance():
    """Return element abundance dictionary."""
    abundance_dict = {}

    for _i, row in _ELEMENT_PRICES.iterrows():
        abundance = row["Abundance in Earth's crust [mg/kg]"]

        if isinstance(abundance, str):
            abundance = abundance.strip("[i]")

            abundance = abundance.strip("~")
            abundance = abundance.strip("≤")
            abundance = abundance.strip()

        abundance_dict[row["Symbol"]] = float(abundance) * 1e-6

    return abundance_dict


_ELEMENT_PRICES_DICT = _get_element_prices()


def _gravimetric_price(element_price_fractions_kg, structure):
    return sum(element_price_fractions_kg.values())


def _volumetric_price(element_price_fractions_kg, structure):
    kg_price = _gravimetric_price(element_price_fractions_kg, structure)
    density = structure.density
    return kg_price * density  # 1 g / cm3 = 1kg / L


def _price_per_atom(element_price_fractions_kg, structure):
    return (
        _gravimetric_price(element_price_fractions_kg, structure)
        * (structure.composition.weight)
        / structure.num_sites
    )


class PriceLowerBound(MOFBaseFeaturizer):
    """Compute a lower bound for the price based on the element prices.

    Obviously, this featurization does well in highlighting rare (and expensive)
    metals.
    It can be useful as additional information in a screening but might
    also be an interesting way to describe the composition of a material
    with one number.

    Price data from `Wikipedia (2020-09-08) <https://en.wikipedia.org/wiki/Prices_of_chemical_elements>`_.

    The units of the projections are:

    - gravimetric: USD / kg
    - volumetric: USD / L
    - per atom: USD / atom

    The implementation is based on some code from `Leopold Talirz <https://ltalirz.github.io/>`_.
    """

    _PROJECTIONS = {
        "gravimetric": _gravimetric_price,
        "volumetric": _volumetric_price,
        "per_atom": _price_per_atom,
    }

    def __init__(
        self, projections: Tuple[str] = ("gravimetric", "volumetric"), primitive: bool = False
    ):
        """Initialize the PriceLowerBound featurizer.

        Args:
            projections (Tuple[str]): List of projections to use.
                Possible values are "gravimetric" and "volumetric".
                Default is ("gravimetric", "volumetric").
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to False.
        """
        self.projections = projections
        self.element_masses = _get_element_abundance()

        self.projections = projections
        super().__init__(primitive=primitive)

    def feature_labels(self) -> List[str]:
        labels = []
        for projection in self.projections:
            labels.append(f"price_lower_bound_{projection}")
        return labels

    def _featurize(self, structure) -> np.ndarray:
        element_masses = {}

        for site in structure.sites:
            element_masses[site.specie.symbol] = site.specie.atomic_mass

        total_mass = sum(element_masses.values())
        # mapping from symbol to price / 1 kg of MOF
        tuples = [
            (symbol, mass / total_mass * _ELEMENT_PRICES_DICT[symbol])
            for symbol, mass in element_masses.items()
        ]
        element_price_fractions_kg = OrderedDict(
            sorted(tuples, key=operator.itemgetter(1), reverse=True)
        )

        features = []
        for projection in self.projections:
            features.append(self._PROJECTIONS[projection](element_price_fractions_kg, structure))

        return np.array(features)

    def citations(self) -> List[str]:
        return []

    def implementors(self) -> List[str]:
        return ["Leopold Talirz", "Kevin Maik Jablonka"]



# -*- coding: utf-8 -*-
"""Test the PH Stats featurizer."""
import pytest

from mofdscribe.featurizers.topology.ph_stats import PHStats

from ..helpers import is_jsonable


def test_ph_stats(hkust_structure, irmof_structure, cof_structure, hkust_la_structure):
    """Ensure we get the correct number of features and that they are different for different structures."""
    for structure in [hkust_structure, irmof_structure, cof_structure]:
        featurizer = PHStats()
        features = featurizer.featurize(structure)
        feature_labels = featurizer.feature_labels()
        assert len(features) == len(feature_labels)

    hkust_feats = featurizer.featurize(hkust_structure)
    irmof_feats = featurizer.featurize(irmof_structure)
    assert (hkust_feats != irmof_feats).any()
    assert is_jsonable(dict(zip(featurizer.feature_labels(), features)))
    assert features.ndim == 1

    # now, try that encoding chemistry with varying atomic radii works
    # we do this by computing the PH histograms for structures with same geometry
    # and connectivity, but different atoms
    featurizer = PHStats(atom_types=None, alpha_weight="atomic_radius_calculated")
    hkust_feats = featurizer.featurize(hkust_structure)
    features_hkust_la = featurizer.featurize(hkust_la_structure)
    assert hkust_feats.shape == features_hkust_la.shape
    assert (hkust_feats != features_hkust_la).any()

    # now, to be sure do not encode the same thing with the atomic radius
    featurizer = PHStats(atom_types=None, alpha_weight=None)
    hkust_feats = featurizer.featurize(hkust_structure)
    features_hkust_la = featurizer.featurize(hkust_la_structure)
    assert hkust_feats.shape == features_hkust_la.shape
    assert hkust_feats == pytest.approx(features_hkust_la, rel=1e-2)



# -*- coding: utf-8 -*-
"""Test APRDF featurizer."""
from mofdscribe.featurizers.chemistry.aprdf import APRDF

from ..helpers import is_jsonable


def test_aprdf(hkust_structure):
    """Make sure that the featurization works for typical MOFs and the number of features is as expected."""
    aprdf_featurizer = APRDF()
    feats = aprdf_featurizer.featurize(hkust_structure)
    assert len(feats) == 432
    assert len(aprdf_featurizer.feature_labels()) == 432
    assert is_jsonable(dict(zip(aprdf_featurizer.feature_labels(), feats)))



# -*- coding: utf-8 -*-
"""Helpers for adverserial validation"""
from __future__ import annotations

from typing import Union

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from sklearn.ensemble import ExtraTreesClassifier, RandomForestClassifier
from sklearn.model_selection import cross_val_score

__all__ = ["AdverserialValidator"]


class AdverserialValidator:
    """Helper for adverserial validation.

    Adverserial is a method to estimate how different two datasets are.
    Most commonly, it is used to estimate if the train and test sets
    come from the same distribution.
    It has found widespread use in data science competions [KaggleBook]_,
    but more recently, also in some auto-ML systems.

    The basic idea is quite simple: Train a classifier to distinguish
    two datasets. If it can learn to do so, there are differences,
    if it cannot, then the datasets are indistinguishable.
    Additionally, this approach also enables us to investigate
    what the most important features for this difference are.
    If one aims to reduce data drift problems, one might remove those
    features [Uber]_.

    Here, we use simple ensemble classifiers such as random forests
    or extra trees to compute the k-fold crossvalidated area under
    the receiver-operating characteristic curve.

    Example:
        >>> from mofdscribe.metrics.adverserial import AdverserialValidator
        >>> x_a = np.array([[1, 2, 3], [4, 5, 6]])
        >>> x_b = np.array([[1, 2, 3], [4, 5, 6]])
        >>> validator = AdverserialValidator(x_a, x_b)
        >>> validator.score().mean()
        0.5

    References:
        .. [Uber] Pan, J.; Pham, V.; Dorairaj, M.; Chen, H.; Lee, J.-Y.
            arXiv:2004.03045 June 26, 2020.

        .. [KaggleBook] Banachewicz, K.; Massaron, L.
            The Kaggle Book: Data Analysis and Machine Learning
            for Competitive Data Science; Packt Publishing, 2022.
    """

    def __init__(
        self,
        x_a: Union[ArrayLike, pd.DataFrame],
        x_b: Union[ArrayLike, pd.DataFrame],
        modeltype: str = "rf",
        k: int = 5,
    ):
        """Initiate a AdverserialValidator instance.

        Args:
            x_a (Union[ArrayLike, pd.DataFrame]): Data for the first dataset (e.g. training set).
            x_b (Union[ArrayLike, pd.DataFrame]): Data for the second dataset (e.g. test set).
            modeltype (str): Classifier to train. Defaults to "rf".
            k (int): Number of folds in k-fold crossvalidation. Defaults to 5.

        Raises:
            ValueError: If the chosen modeltype is not supported.
        """
        if modeltype == "rf":
            self.model = RandomForestClassifier()
        elif modeltype == "et":
            self.model = ExtraTreesClassifier()
        else:
            raise ValueError(f"Model {modeltype} not implements. Available models are rf, et.")

        self.x_a = x_a
        self.x_b = x_b
        self.k = k

    def _get_x_y(self):
        if isinstance(self.x_a, pd.DataFrame):
            x = pd.concat([self.x_a, self.x_b])

        else:
            x = np.vstack([self.x_a, self.x_b])

        y = [0] * len(self.x_a) + [1] * len(self.x_b)

        return x, y

    def score(self) -> np.array:
        """Compute the area under the receiver-operating characteristic curve.

        A score close to 0.5 means that the two datasets are similar.

        Returns:
            np.array: Areas under the receiver-operating characteristic curve.

        """
        x, y = self._get_x_y()
        score = cross_val_score(self.model, x, y, scoring="roc_auc")
        return score

    def get_feature_importance(self) -> np.array:
        """Identify the features distinguishing the two datasets.

        Uses the default impurity-based feature importance.

        Returns:
            np.array: Feature importance scores.
        """
        x, y = self._get_x_y()
        self.model.fit(x, y)

        importances = self.model.feature_importances_
        return importances



# -*- coding: utf-8 -*-
"""perform analyses on structure graphs."""
from functools import lru_cache
from typing import List, Optional, Set, Tuple

from pymatgen.analysis.graphs import StructureGraph
from pymatgen.core import IStructure
from structuregraph_helpers.create import get_structure_graph as get_sg


def get_neighbors_at_distance(
    structure_graph: StructureGraph, start: int, scope: int
) -> Tuple[Set[int], List[int]]:
    """For a structure graph and a start site, return all sites within a certain\
        distance (scope) of the start site.

    Args:
        structure_graph (StructureGraph): pymatgen StructureGraph object
        start (int): starting atom
        scope (int): distance to search

    Returns:
        Tuple[Set[int], List[int]]: All sites within the scope of the start
        site, and the indices of the sites in the last shell
    """
    # Todo: This code is stupid.
    neighbors_at_last_level = [start]
    all_neighbors = set()
    neighbors_at_next_level = []
    for _ in range(scope):
        for n in neighbors_at_last_level:
            neighbors_at_next_level.extend(get_connected_site_indices(structure_graph, n))

        all_neighbors.update(neighbors_at_last_level)
        neighbors_at_last_level = neighbors_at_next_level
        neighbors_at_next_level = []
    all_neighbors.remove(start)
    neighbors_at_last_level = set(neighbors_at_last_level)
    if start in neighbors_at_last_level:
        neighbors_at_last_level.remove(start)

    return all_neighbors, neighbors_at_last_level


@lru_cache()
def get_structure_graph(structure: IStructure, strategy: Optional[str] = None) -> StructureGraph:
    """Get a StructureGraph object for a given structure.

    Args:
        structure (IStructure): Pymatgen structure object.
        strategy (str): Heuristic for assigning bonds.
            Must be one of 'jmolnn', 'crystalnn', 'isayevnn', 'vesta'. Defaults to None.

    Returns:
        StructureGraph: pymatgen StructureGraph object
    """
    return get_sg(structure, strategy)


def get_connected_site_indices(structure_graph: StructureGraph, site_index: int) -> List[int]:
    """Get all connected site indices for a given site."""
    connected_sites = structure_graph.get_connected_sites(site_index)
    return [site.index for site in connected_sites]
